from . import BolometerProperties
from .. import core
import numpy as np
import re

__all__ = [
    "SplitByProperty",
    "SplitByBand",
    "SplitTimestreamsByBand",
    "SplitByWafer",
    "SplitByPixelType",
    "BandFormat",
    "get_band_units",
    "set_band_format",
    "band_to_string",
    "band_to_value",
    "extract_band_string",
    "extract_band_value",
]


# framework for handling frequency band formatting

class BandFormat:
    """
    Class for converting a frequency band between a quantity in G3Units and its
    string representation.

    Arguments
    ---------
    precision : int
        Float formatting precision for the band quantity.  If <=0, output will
        be an integer.  If >0, output will be a floating point number with this
        many decimal places of precision.
    units : str
        String name of the G3Units quantity in which to represent the frequency
        band, e.g. "GHz" or "MHz".  Must correspond to a valid attribute of the
        core.G3Units namespace.
    """

    def __init__(self, precision=0, units="GHz"):
        self.set_format(precision, units)

    def set_format(self, precision=0, units="GHz"):
        """
        Set the band format precision and units for converting between a
        quantity in G3Units and its string representation.

        Arguments
        ---------
        precision : int
            Float formatting precision for the band quantity.  If <=0, output
            will be an integer.  If >0, output will be a floating point number
            with this many decimal places of precision.
        units : str
            String name of the G3Units quantity in which to represent the
            frequency band, e.g. "GHz" or "MHz".  Must correspond to a valid
            attribute of the core.G3Units namespace.
        """
        self._precision = int(precision)
        assert hasattr(core.G3Units, units), "Invalid units {}".format(units)
        self._uvalue = getattr(core.G3Units, units)
        self._vformat = "%%.%df" % (precision if precision > 0 else 0)
        self._uformat = units
        self._format = "%s%s" % (self._vformat, units)
        prx = r"\.[0-9]{%d}" % precision if precision > 0 else ""
        self._pattern = "([0-9]+%s)%s" % (prx, units)
        self._wregex = re.compile("^" + self._pattern + "$")
        self._regex = re.compile(self._pattern)

    @property
    def units(self):
        """Units string used for formatting"""
        return self._uformat

    def to_string(self, value, include_units=True):
        """Convert a band value in G3Units to its string representation, using
        the appropriate precision and units name."""
        if not np.isfinite(value) or value < 0:
            return ""
        value = np.round(value / self._uvalue, self._precision)
        if not include_units:
            return self._vformat % value
        return self._format % value

    def to_value(self, string):
        """Convert a band string to a value in G3Units, or raise a ValueError
        if no match is found."""
        m = self._wregex.match(string)
        if not m:
            raise ValueError("Invalid band {}".format(string))
        return float(m.group(1)) * self._uvalue

    def extract_string(self, string, include_units=True):
        """Return the band substring from the input string, or None if not
        found."""
        m = self._regex.match(string)
        if not m:
            return None
        if not include_units:
            return m.group(1)
        return m.group(0)

    def extract_value(self, value):
        """Return the band in G3Units extracted from the input string, or None
        if not found."""
        s = extract_string(value, include_units=False)
        if not s:
            return None
        return float(v) * self._uvalue


# global instance used by functions below
_band_format = BandFormat()


@core.usefulfunc
def get_band_units():
    """Return the units string used for formatting frequency band values."""
    return _band_format.units


@core.usefulfunc
def set_band_format(precision, units):
    _band_format.set_format(precision, units)
set_band_format.__doc__ = BandFormat.set_format.__doc__


@core.usefulfunc
def band_to_string(value, include_units=True):
    return _band_format.to_string(value, include_units=include_units)
band_to_string.__doc__ = BandFormat.to_string.__doc__


@core.usefulfunc
def band_to_value(string):
    return _band_format.to_value(string)
band_to_value.__doc__ = BandFormat.to_value.__doc__


@core.usefulfunc
def extract_band_string(string, include_units=True):
    return _band_format.extract_string(string, include_units=include_units)
extract_band_string.__doc__ = BandFormat.extract_string.__doc__


@core.usefulfunc
def extract_band_value(string):
    return _band_format.extract_value(string)
extract_band_value.__doc__ = BandFormat.extract_value.__doc__


# monkeypatch useful property attributes
def band_string(self):
    """String representation of frequency band center"""
    return _band_format.to_string(self.band)
BolometerProperties.band_string = property(band_string)

def band_vstring(self):
    """String representation of frequency band center, without units name"""
    return _band_format.to_string(self.band, include_units=False)
BolometerProperties.band_vstring = property(band_vstring)


@core.indexmod
class SplitByProperty(object):
    '''
    Take an input G3FrameObject-derivative Map keyed by bolometer name and
    split it into several based on the property of the detectors as given by
    the BolometerProperties key.
    Return the same type of maps as the one it was handed, e.g.
    G3TimestreamMap, G3MapInt, etc.
    '''
    def __init__(self, input='CalTimestreams', property=None, property_list=None,
                 output_root=None, bpm='BolometerProperties', drop_empty=False):
        '''
        Split the input map given by input into several output
        maps named output_root + key (e.g. CalTimestreams + str(property)) with
        the default options).

        Arguments
        ---------
        input : str
            Key name of the input map to split.
        property : str
            Attribute name to extract from the BolometerProperties object.
            Required.
        property_list : list of properties
            Properties to include in the output keys.  Entries that are not strings
            will be converted to strings using the `SplitByProperty.converter` method.
            If property_list is not None, use only the names in the
            list (possibly writing empty timestream maps to the frame). Otherwise,
            creates maps for every that exists in the input.
        output_root : str
            Prefix for the output keys.
            If None (default), use `input` as the output root.
        bpm : str
            The key name of the BolometerPropertiesMap from which to extract
            the requested `property` for splitting the input map.
        drop_empty : bool
            If True, drop output maps that don't contain any bolometers.
        '''
        if property is None:
            core.log_fatal("Property is a required argument")
        self.bpmattr = property

        self.input = input
        self.output_root = output_root if output_root is not None else input
        if property_list is not None:
            self.props = [self.converter(x) if not isinstance(x, str) else x
                          for x in property_list]
        else:
            self.props = None
        self.bpmkey = bpm
        self.bpm = None
        self.drop_empty = drop_empty

    @staticmethod
    def converter(prop):
        """
        Function for converting the property to its corresponding string name.
        Returns a string representation of the input argument, or None if the
        argument is invalid.

        Overload this function in subclasses of SplitByProperty to change
        how attributes are parsed into their string representations.
        """
        if prop is None:
            return None
        return str(prop)

    def __call__(self, frame):
        if self.bpmkey in frame:
            self.bpm = frame[self.bpmkey]

        if self.input not in frame:
            return

        inmap = frame[self.input]
        out = {}
        if self.props is not None:
            for prop in self.props:
                out[prop] = type(inmap)()

        for b in inmap.keys():
            try:
                prop = self.converter(getattr(self.bpm[b], self.bpmattr))
            except KeyError:
                continue
            if prop not in out:
                if self.props is None and prop is not None:
                    out[prop] = type(inmap)()
                else:
                    continue
            out[prop][b] = inmap[b]

        for prop, outmap in out.items():
            if not len(outmap.keys()) and self.drop_empty:
                continue
            frame['%s%s' % (self.output_root, prop)] = outmap


@core.indexmod
class SplitByBand(SplitByProperty):
    '''
    Take an input G3FrameObject-derivative Map keyed by bolometer name and
    split it into several based on the bands of the detectors as given by
    the BolometerProperties key.
    Return the same type of maps as the one it was handed, e.g.
    G3TimestreamMap, G3MapInt, etc.
    '''
    def __init__(self, input='CalTimestreams', output_root=None,
                 bands=None, bpm='BolometerProperties', drop_empty=False):
        '''
        Split the input map given by input into several output
        maps named output_root + band + GHz (e.g. CalTimestreams150GHz with
        the default options). If bands is not None, use only the bands in the
        list (possibly writing empty timestream maps to the frame). Otherwise,
        creates maps for every band that exists in the input. Setting bpm
        to a non-default value causes this to get its band mapping from an
        alternative data source.
        '''
        super(SplitByBand, self).__init__(
            input=input, output_root=output_root, property_list=bands,
            bpm=bpm, property='band_string', drop_empty=drop_empty)

    @staticmethod
    def converter(band_string):
        if band_string is None:
            return None
        if not band_string:
            return None
        return str(band_string)


@core.indexmod
class SplitTimestreamsByBand(SplitByBand):

    def __init__(self, input='CalTimestreams', output_root=None,
                 bands=None, bpm='BolometerProperties', drop_empty=False):
        core.log_warn("SplitTimestreamsByBand is deprecated, use SplitByBand instead")
        super(SplitTimestreamsByBand, self).__init__(
            input=input, output_root=output_root, bands=bands, bpm=bpm, drop_empty=drop_empty)


@core.indexmod
class SplitByWafer(SplitByProperty):
    '''
    Take an input G3FrameObject-derivative Map keyed by bolometer name and
    split it into several based on the wafers of the detectors as given by
    the BolometerProperties key.
    Return the same type of maps as the one it was handed, e.g.
    G3TimestreamMap, G3MapInt, etc.
    '''
    def __init__(self, input='CalTimestreams', output_root=None,
                 wafers=None, bpm='BolometerProperties', drop_empty=False):
        '''
        Split the input map given by input into several output
        maps named output_root + wafer (e.g. CalTimestreamsW172 with
        the default options). If wafers is not None, use only the wafers in the
        list (possibly writing empty timestream maps to the frame). Otherwise,
        creates maps for every wafer that exists in the input. Setting bpm
        to a non-default value causes this to get its wafer mapping from an
        alternative data source.
        '''
        super(SplitByWafer, self).__init__(
            input=input, output_root=output_root, property_list=wafers,
            bpm=bpm, property='wafer_id', drop_empty=drop_empty)

    @staticmethod
    def converter(wafer):
        if wafer is None:
            return None
        return str(wafer).capitalize()


@core.indexmod
class SplitByPixelType(SplitByProperty):
    '''
    Take an input G3FrameObject-derivative Map keyed by bolometer name and
    split it into several based on the pixel types of the detectors as given by
    the BolometerProperties key.
    Return the same type of maps as the one it was handed, e.g.
    G3TimestreamMap, G3MapInt, etc.
    '''
    def __init__(self, input='CalTimestreams', output_root=None,
                 types=None, bpm='BolometerProperties', drop_empty=False):
        '''
        Split the input map given by input into several output
        maps named output_root + wafer (e.g. CalTimestreamsW172 with
        the default options). If wafers is not None, use only the wafers in the
        list (possibly writing empty timestream maps to the frame). Otherwise,
        creates maps for every wafer that exists in the input. Setting bpm
        to a non-default value causes this to get its wafer mapping from an
        alternative data source.
        '''
        super(SplitByPixelType, self).__init__(
            input=input, output_root=output_root, property_list=types,
            bpm=bpm, property='pixel_type', drop_empty=drop_empty)

    @staticmethod
    def converter(pixel_type):
        if pixel_type is None:
            return None
        if not pixel_type:
            return None
        pixel_type = str(pixel_type)
        if pixel_type.lower() == 'n/a':
            return None
        if pixel_type.islower():
            return pixel_type.capitalize()
        return pixel_type



import spt3g._libdfmux
for k in dir(spt3g._libdfmux):
    print(k, getattr(spt3g._libdfmux, k))
from spt3g._libdfmux import DfMuxSample
import spt3g.dfmux
from spt3g import dfmux
from spt3g.dfmux import DfMuxSample



#!/usr/bin/env python

from spt3g import core
import numpy as np

dtypes = [np.int32, np.int64, np.float32, np.float64]
start = core.G3Time.Now()
stop = start + 5 * core.G3Units.s
ndet = 20
nsamp = 1200
keys = [str(x).zfill(3) for x in range(ndet)]

for bit_depth in [24, 32]:
    for dtype in dtypes:
        print("bit depth", bit_depth, "dtype", dtype)
        dat = np.random.normal(size=(ndet, nsamp), scale=2**28, loc=0).astype(dtype)
        if dtype in [np.float32, np.float64]:
            dat[:, 5] = np.nan
        tsm = core.G3TimestreamMap(
            keys, dat, start=start, stop=stop, compression_level=5, bit_depth=bit_depth
        )
        try:
            with core.G3Writer("test.g3") as w:
                fr = core.G3Frame()
                fr["tsm"] = tsm
                w(fr)
        except Exception as e:
            if "32-bit compression is not supported" in str(e):
                continue
            raise

        tsm2 = list(core.G3File("test.g3"))[0]["tsm"]

        assert tsm.start == tsm2.start
        assert tsm.stop == tsm2.stop
        assert tsm.compression_level == tsm2.compression_level
        assert tsm.bit_depth == tsm2.bit_depth

        if bit_depth == 24 and tsm.dtype == np.float64:
            assert tsm2.dtype == np.float32
        elif bit_depth == 24 and tsm.dtype == np.int64:
            assert tsm2.dtype == np.int32
        else:
            assert tsm.dtype == tsm2.dtype

        if dtype in [np.float32, np.float64]:
            assert np.isnan(np.asarray(tsm2)[:, 5]).all()

        def mask32(v):
            return (np.asarray(v)[:, 6:]).astype(np.int32)

        tsm_trunc = mask32(tsm)
        if bit_depth == 24:
            tsm_trunc = ((tsm_trunc & 0x00FFFFFF) << 8) >> 8
        np.testing.assert_array_equal(tsm_trunc, mask32(tsm2))



#!/usr/bin/env python

from spt3g import core

import unittest
import os

test_filename = 'inttest.g3'

# Type of int needed to store the specified value.
bit_sizes = [
    ( 8,  0),
    ( 8,  1),
    ( 8, -1),
    ( 8,  127),
    ( 8, -128),
    (16,  128),
    (16, -129),
    (16,  32767),
    (16, -32768),
    (32,  32768),
    (32, -32769),
    (32,  2147483647),
    (32, -2147483648),
    (64,  2147483648),
    (64,  2147483649),
    (64,  9223372036854775807),
    (64, -9223372036854775808),
]

class TestVectorInt(unittest.TestCase):
    def tearDown(self):
        os.remove(test_filename)
        
    def test_serialize(self):
        """Confirm full ranges can be saved and loaded."""

        with core.G3Writer(test_filename) as w:
            for isize, val in bit_sizes:
                f = core.G3Frame()
                f['v'] = core.G3VectorInt([val] * 10)
                w(f)

        r = core.G3Reader(test_filename)
        for isize, val in bit_sizes:
            f = r(None)[0]
            v_in = list(f['v'])
            self.assertTrue(all([_v == val for _v in v_in]),
                             "Failed to save/load value %i" % val)
        del r

    def test_serialize_map(self):
        """Confirm full ranges can be saved and loaded."""

        with core.G3Writer(test_filename) as w:
            for isize, val in bit_sizes:
                f = core.G3Frame()
                f['m'] = core.G3MapInt({str(i): val for i in range(10)})
                w(f)

        r = core.G3Reader(test_filename)
        for isize, val in bit_sizes:
            f = r(None)[0]
            v_in = list(f['m'].values())
            self.assertTrue(all([_v == val for _v in v_in]),
                             "Failed to save/load value %i" % val)
        del r

    def test_serialize_map_vector(self):
        """Confirm full ranges can be saved and loaded."""

        with core.G3Writer(test_filename) as w:
            for isize, val in bit_sizes:
                f = core.G3Frame()
                f['m'] = core.G3MapVectorInt({'a': [val] * 10})
                w(f)

        r = core.G3Reader(test_filename)
        for isize, val in bit_sizes:
            f = r(None)[0]
            v_in = list(f['m']['a'])
            self.assertTrue(all([_v == val for _v in v_in]),
                             "Failed to save/load value %i" % val)
        del r

    def test_compression(self):
        """Confirm that minimum necessary int size is used for serialization."""
        count = 10000
        overhead = 200
        for isize, val in bit_sizes:
            with core.G3Writer(test_filename) as w:
                f = core.G3Frame()
                f['v'] = core.G3VectorInt([val] * count)
                w(f)
            on_disk = os.path.getsize(test_filename)
            self.assertTrue(abs(on_disk - count * isize / 8) <= overhead,
                            "Storage for val %i took %.2f bytes/item, "
                            "too far from %.2f bytes/item" %
                            (val, on_disk / count, isize / 8))

    def test_compression_map(self):
        """Confirm that minimum necessary int size is used for serialization."""
        count = 10000
        overhead = 12
        for isize, val in bit_sizes:
            with core.G3Writer(test_filename) as w:
                f = core.G3Frame()
                f['v'] = core.G3MapInt({str(i): val for i in range(count)})
                w(f)
            on_disk = os.path.getsize(test_filename)
            self.assertTrue(abs(on_disk - count * isize / 8) <= (overhead * count),
                            "Storage for val %i took %.2f bytes/item, "
                            "too far from %.2f bytes/item" %
                            (val, on_disk / count - overhead, isize / 8))

    def test_compression_map_vector(self):
        """Confirm that minimum necessary int size is used for serialization."""
        count = 10000
        overhead = 200
        for isize, val in bit_sizes:
            with core.G3Writer(test_filename) as w:
                f = core.G3Frame()
                f['v'] = core.G3MapVectorInt({'a': [val] * count})
                w(f)
            on_disk = os.path.getsize(test_filename)
            self.assertTrue(abs(on_disk - count * isize / 8) <= overhead,
                            "Storage for val %i took %.2f bytes/item, "
                            "too far from %.2f bytes/item" %
                            (val, on_disk / count, isize / 8))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from spt3g import core
import numpy
import pickle

f64 = numpy.ones(20,dtype='float64')
f32 = numpy.ones(20,dtype='float32')
i32 = numpy.ones(20,dtype='int32')
i64 = numpy.ones(20,dtype='int64')
i16 = numpy.ones(20,dtype='int16') # This one shouldn't work directly

for d in [f64, f32, i32, i64, i16]:
	print(d.dtype)
	try:
		t = core.G3Timestream(d)
	except TypeError:
		if d.dtype == 'int16': # Meant to fail
			continue
		raise
	else:
		if d.dtype == 'int16': # Meant to fail
			raise TypeError('int16 should not work')


	assert(t.dtype == d.dtype)
	assert(len(t) == 20)
	assert(t[12] == 1.0)
	t[12] = 5
	assert(t[12] == 5.0)
	assert(t[13] == 1.0)
	d2 = numpy.asarray(t)
	assert(d2.dtype == d.dtype)
	assert(not (d2 == d).all())
	
	# Test shared buffer, both ways
	t[12] = 1
	assert((d2 == d).all())
	d2[12] = 3
	assert(t[12] == 3)

	# Test serialization
	buf = pickle.dumps(t)
	t2 = pickle.loads(buf)
	assert(t2.dtype == d.dtype) # Datatype preserved
	assert((numpy.asarray(t2) == d2).all())

	# Test copy constructor
	t2 = core.G3Timestream(t)
	assert(t2.dtype == d.dtype) # Datatype preserved
	assert((numpy.asarray(t2) == d2).all())

	# Test map
	m = core.G3TimestreamMap(['a', 'b', 'c'], numpy.asarray([d, d, d]))
	assert(m.dtype == d.dtype)

	buf = pickle.dumps(m)
	m2 = pickle.loads(buf)
	assert(m2.dtype == d.dtype) # Datatype preserved
	assert((numpy.asarray(m2) == numpy.asarray(m)).all())

	m2 = core.G3TimestreamMap(m)
	assert(m2.dtype == d.dtype) # Datatype preserved
	assert((numpy.asarray(m2) == numpy.asarray(m)).all())



from .. import core
import os, time
import numpy as np

@core.usefulfunc
class CalFileReader(object):
    '''
    For now just reads a G3 calibration file and loads the contents into 
    a dictionary
    '''

    def __init__(self, calibration_file=None):
        self.default_calibration_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'G3RegCal.txt')

    def readCalFile(self, calfile=None):
        cal_dict = {}
        if calfile is None:
            self.calibration_file = self.default_calibration_file
        else:
            self.calibration_file = calfile
        try:
            f = open(self.calibration_file, 'r')
        except:
            core.log_warn('3G calibration file ' + 
                              self.calibration_file + ' not found.\n')
            return cal_dict
        ninvalid = 0
        for line in f:
            if line[0] != '#' and line[0] != '\n' and len(line) > 0:
                try:
                    line = line.replace('\n','')
                    fline = list(filter(None,line.split(' '))) #list() for py3
                    name = fline[0]
                    linedict = {}
                    linedict['Offset'] = float(fline[1])
                    linedict['ReciprocalFactor'] = float(fline[2])
                    try:
                        linedict['UnitName'] = fline[3]
                    except:
                        linedict['UnitName'] = 'None'
                    cal_dict[name] = linedict
                except:
                    ninvalid += 1
        if ninvalid > 0:
            if ninvalid == 1:
                core.log_warn('One invalid line found and ignored in calibration file.\n')
            else:
                core.log_warn(str(ninvalid) + ' invalid lines found and ignored in calibration file.\n')
        f.close()

        # now parse into dictionary structure NW uses in ARCExtractor
        cal_dict_orig = cal_dict.copy()
        cal_dict = {}
        origkeys = cal_dict_orig.keys()
        origkeys = sorted(origkeys)
        names = [key for key in origkeys]
        regmaps = [(list(filter(None,(tname).split('.'))))[0] for tname in names]
        uregmaps = np.unique(np.asarray(regmaps))
        for urm in uregmaps:
            cal_dict[urm] = {}
        regblocks = [(list(filter(None,(tname).split('.'))))[1] for tname in names]
        uregblocks,argurb = np.unique(np.asarray(regblocks),return_index=True)
        for j in np.arange(len(uregblocks)):
            cal_dict[regmaps[argurb[j]]][uregblocks[j]] = {}
        regs = [(list(filter(None,(tname).split('.'))))[2] for tname in names]
        for j in np.arange(len(regs)):
            reg = regs[j]
            if '[' in reg:
                reg_plus_indices = reg.split('[')
                realreg = reg_plus_indices[0]
                indices = ((reg_plus_indices[1]).split(']'))[0]
                if '-' in indices:
                    cal_dict[regmaps[j]][regblocks[j]][realreg] = {}
                    siindices = indices.split('-')
                    iindices = (np.array(siindices)).astype(int)
                    for i in np.arange(np.max(iindices)-np.min(iindices)+1) + \
                            np.min(iindices):
                        cal_dict[regmaps[j]][regblocks[j]]\
                            [realreg][i] = cal_dict_orig[names[j]]
                else:
                    if indices == '0':
                        cal_dict[regmaps[j]][regblocks[j]][realreg] = {}
                    cal_dict[regmaps[j]][regblocks[j]][realreg]\
                        [int(indices)] = cal_dict_orig[names[j]]
            else:
                cal_dict[regmaps[j]][regblocks[j]][reg] = \
                    cal_dict_orig[names[j]]

        return cal_dict

@core.usefulfunc
def create_g3_cal_file(path, read_from_gcp=True, extra_dict=None,
                       use_extra_info=True, gcp_cal_file=None):
    '''
    Create a G3 register calibration file. Usually reads in
    calibration and units from a GCP cal file then adds extra
    information that isn't in the GCP one. This extra information can
    be handed to the routine as a dict (note format below); otherwise,
    the hard-coded extra info will be used. (Feel free to edit this
    hard-coded info.)
    '''

    cal_dict = {}
    status = 0
    
    if read_from_gcp:
        try:
            default_calibration_file = \
                os.environ['GCP_DIR']+'/config/init/cal'
        except:
            default_calibration_file = \
                '/home/sptdat/gcproot/gcp/config/init/cal'
        if gcp_cal_file is None:
            calibration_file = default_calibration_file
        else:
            calibration_file = gcp_cal_file
        try:
            f = open(calibration_file)
        except:
            core.log_warn('GCP calibration file ' + 
                          calibration_file + ' not found.\n')
            f = []
        ninvalid = 0
        for line in f:
            if line[0] != '#' and line[0] != '\n' and len(line) > 0:
                line = line.replace('\n','')
                linedict = {}
                # cal file has spaces AND tabs (and comment characters), whee!
                info_and_comment = list(filter(None,line.split('#')))
                fline = list(filter(None,(info_and_comment[0]).split(' ')))
                if len(fline) == 1 and '\t' in fline[0]:
                    fline = list(filter(None,(info_and_comment[0]).split('\t')))
                name = (fline[0]).replace('*','0')
                try:
                    linedict['Offset'] = float(fline[1])
                    if '/' in fline[2]:
                        gainfacs = (fline[2]).split('/')
                        gainfac = float(gainfacs[1])/float(gainfacs[0])
                    else:
                        gainfac = 1./float(fline[2])
                    linedict['ReciprocalFactor'] = gainfac
                    linedict['UnitName'] = ''
                # try to figure out units. currently cal file 
                # only has 2 forms, but this is not robust.
                    if len(info_and_comment) > 1: 
                        comment = info_and_comment[1]
                        if '->' in comment:
                            comments = list(filter(None,(comment).split('->')))
                            unitname = \
                                list(filter(None,(comments[1]).split(' ')))[0]
                            unitname = list(filter(None,(unitname).split('/')))[0]
                        else:
                            comments = \
                                list(filter(None,(comment).split('display in ')))
                            unitname = \
                                list(filter(None,(comments[0]).split(' ')))[0]
                            unitname = unitname.replace('\n','')
                        linedict['UnitName'] = unitname
                # another hudge kludge
                        if 'rate' in name:
                            if '/' not in unitname:
                                unitname = unitname + '/s'
                                linedict['UnitName'] = unitname
                    cal_dict[name] = linedict
                except:
                    ninvalid += 1
        if ninvalid > 0:
            if ninvalid == 1:
                core.log_warn('One invalid line found and ignored in calibration file.\n')
            else:
                core.log_warn(str(ninvalid) + ' invalid lines found and ignored in calibration file.\n')
        f.close()

    # now add extra info
    if use_extra_info:
        if extra_dict is None:
            cal_dict['antenna0.acu.az_pos'] = \
                {'ReciprocalFactor':1.,'Offset':0,'UnitName':'deg'}
            cal_dict['antenna0.acu.el_pos'] = \
                {'ReciprocalFactor':1.,'Offset':0,'UnitName':'deg'}
            cal_dict['antenna0.acu.az_err'] = \
                {'ReciprocalFactor':1.,'Offset':0,'UnitName':'deg'}
            cal_dict['antenna0.acu.el_err'] = \
                {'ReciprocalFactor':1.,'Offset':0,'UnitName':'deg'}

            cal_dict['antenna0.scu.temp'] = \
                {'ReciprocalFactor':1.,'Offset':273.15,'UnitName':'K'}
            cal_dict['array.weather.airTemperature'] = \
                {'ReciprocalFactor':1.,'Offset':273.15,'UnitName':'K'}
        else:
            try:
                for key in extra_dict:
                    cal_dict[key] = {}
                    cal_dict[key]['Offset'] = extra_dict[key]['Offset']
                    cal_dict[key]['ReciprocalFactor'] = \
                        extra_dict[key]['ReciprocalFactor']
                    cal_dict[key]['UnitName'] = extra_dict[key]['UnitName']
            except:
                core.log_warn('extra_dict input format not recognized, ignoring')
                del cal_dict[key]

    # write cal file
    fcal = open(path,'w')
    fcal.write('#-----------------------------------------------------------------------\n')
    fcal.write('# This is an SPT-3G register calibration file. \n')
    fcal.write('#\n')
    fcal.write('# It was created on ' + time.ctime() + ' by user ' + os.getlogin() + '\n')
    fcal.write('#  with the spt3g_software function gcp.CalFile.create_g3_cal_file. \n')
    fcal.write('#\n')
    fcal.write('# Each line has the following format:\n')
    fcal.write('#\n')
    fcal.write('#   register offset 1/factor units\n')
    fcal.write('#\n')
    fcal.write('# (Note that factors are stored as reciprocals, because they are most \n')
    fcal.write('#  often one over a whole number.) \n')
    fcal.write('#\n')
    fcal.write('# The calibrated value of a register is given by:\n')
    fcal.write('#\n')
    fcal.write('#   reg_cal = factor * (offset + register_value) * core.G3Units.$UNIT$\n')
    fcal.write('#\n')
    fcal.write('#-----------------------------------------------------------------------\n')
    fcal.write('# Register      Offset       1/Factor      Units \n')
    fcal.write('#-----------------------------------------------------------------------\n')
    fcal.write('\n')
    ktemp = list(cal_dict.keys())
    ktemp.sort()
    for key in ktemp:
        kstr = (str(key)).ljust(42)
        gstr = '%e' % cal_dict[key]['ReciprocalFactor']
        ostr = '%10.4f' % cal_dict[key]['Offset']
        fcal.write(kstr + '  ' + ostr + '  ' + gstr + '  ' + cal_dict[key]['UnitName'] + '\n')
    fcal.close()

    return status




#!/usr/bin/env python
from spt3g import core
import numpy as np

a = core.Quat(2,3,4,5)

assert(a.real == 2)
assert(a.real == np.real(a))
assert(a+a == 2*a)
assert(a+a == a*2)
assert(a*a == a**2)
assert(a*a*a == a**3)

b = core.G3VectorQuat([a, a**2, 2*a])

assert(b[0].a == 2)
assert(b[0].b == 3)
assert(b[0].c == 4)
assert(b[0].d == 5)

assert(b[1].a == -46)
assert(b[1].b == 12)
assert(b[1].c == 16)
assert(b[1].d == 20)

assert(b[2].c == 8)

assert((b.real == np.asarray([2, -46, 4])).all())
assert((b.real == np.real(b)).all())

c = np.asarray(b)

assert(c.shape == (3,4))

assert(core.Quat(*c[0]) == a)
assert(core.Quat(*c[1]) == b[1])
assert(core.Quat(*c[1]) != b[2])

d = core.G3VectorQuat(c)

assert(d[0] == b[0])
assert(d[1] == b[1])
assert(d[1] != b[2])
assert(d[2] == b[2])

c[1,3] = 15
assert(b[1].d == 15)

e = 2*b
assert((2*np.asarray(b) == np.asarray(e)).all())
e = b/2
assert((np.asarray(b)/2 == np.asarray(e)).all())

assert((np.asarray(b*b) == np.asarray(core.G3VectorQuat([x**2 for x in b]))).all())

assert(1./a == core.Quat(1.,0,0,0)/a)
assert(np.allclose(core.G3VectorQuat([1./a]), core.G3VectorQuat([(~a) / abs(a)**2])))
assert(len(np.abs(b)) == len(b))
assert((np.abs(b) == np.asarray(abs(b))).all())
assert(((np.abs(b) - np.abs(~b)) == 0).all())
assert(a/b[0] == (a/b)[0])
assert(b[1]/a == (b/a)[1])

# Test for support of numpy slicing and conversions
quats = np.array([[1., 2., 3., 4., 5.],   # a
                  [0., 0., 0., 3., -1.],   # b
                  [0., 0., 0., 0., 0.],   # c
                  [18., -23., 5., 0., 0.]])  # d

# numpy slicing and conversions of single quaternions
assert(core.Quat(quats[0, :4]) == core.Quat(*quats[0, :4]))
assert((quats[0, :4] == np.asarray(core.Quat(*quats[0, :4]))).all())
assert(core.Quat(quats[:, 0]) == core.Quat(*quats[:, 0]))
assert((quats[:, 0] == np.asarray(core.Quat(*quats[:, 0]))).all())

try:
	q = core.G3VectorQuat(quats)
except TypeError:
	# Wrong shape, should fail
	pass
else:
	raise TypeError('Wrong shape, should not convert')

# Non-trivial strides
q = core.G3VectorQuat(quats[:,:4])
assert(q[0] == core.Quat(*quats[0,:4]))
assert(q[1] == core.Quat(*quats[1,:4]))
assert(q[2] == core.Quat(*quats[2,:4]))
assert(q[3] == core.Quat(*quats[3,:4]))

# When transposed, has right shape to convert
q = core.G3VectorQuat(quats.T) # Strides, but simple ones
assert(q[0] == core.Quat(*quats[:, 0]))
assert(q[1] == core.Quat(*quats[:, 1]))
assert(q[2] == core.Quat(*quats[:, 2]))
assert(q[3] == core.Quat(*quats[:, 3]))
assert(q[4] == core.Quat(*quats[:, 4]))

# Trivial case, packed
q = core.G3VectorQuat(quats.T.copy())
assert(q[0] == core.Quat(*quats[:, 0]))
assert(q[1] == core.Quat(*quats[:, 1]))
assert(q[2] == core.Quat(*quats[:, 2]))
assert(q[3] == core.Quat(*quats[:, 3]))
assert(q[4] == core.Quat(*quats[:, 4]))

# Test conversion of integers

qint = np.asarray(quats[:,:4], dtype='int64')
q = core.G3VectorQuat(qint)
assert(q[0] == core.Quat(1,2,3,4))
assert(q[1] == core.Quat(0,0,0,3))
assert(q[3] == core.Quat(18, -23, 5, 0))

# Test serialization

frame = core.G3Frame()
q = core.Quat(1, 2, 3, 4)
frame["Quat"] = q
assert frame["Quat"] == q



#!/usr/bin/env python

import numpy as np
from spt3g import core
from spt3g.maps import G3SkyMapWeights, FlatSkyMap, MapPolType, MapPolConv, MapProjection
from spt3g.maps import remove_weights, remove_weights_t, apply_weights, apply_weights_t, flatten_pol

for pol in [True, False]:
    # allocation
    pol_conv = MapPolConv.IAU if pol else MapPolConv.none
    m = FlatSkyMap(500, 20, core.G3Units.arcmin, pol_conv=pol_conv, proj=MapProjection.Proj5)
    mt = m.clone(False)
    mt.pol_type = MapPolType.T
    if pol:
        mq = m.clone(False)
        mq.pol_type = MapPolType.Q
        mu = m.clone(False)
        mu.pol_type = MapPolType.U
    mw = G3SkyMapWeights(m)
    assert(mw.size == m.size)
    assert(mw.shape == m.shape)
    assert(mw.npix_allocated == 0)
    assert(mw.polarized == pol)
    assert(mw.congruent)

    assert(mw.compatible(m))
    for k in mw.keys():
        assert(mw[k].compatible(mw.TT))

    if not pol:
        assert(mw.QQ is None)
        assert(mw.UU is None)

    # assignment
    vec = np.array([10., 1., 1.]) if pol else 10.
    if pol:
        mt[15], mq[15], mu[15] = vec
        assert(np.allclose([mt[15], mq[15], mu[15]], vec))
    else:
        mt[15] = vec
        assert(np.allclose(mt[15], vec))

    mat = np.array([[2, 0.2, 0.1],
                    [0.2, 0.5, 0.05],
                    [0.1, 0.05, 0.25]]) if pol else 2.
    mw[15] = mat
    assert(mw.npix_allocated == 1)
    assert(np.allclose(mw[15], mat))

    # operators
    mw += 2 * mw
    assert(np.allclose(mw[15], mat * 3))
    assert(mw.npix_allocated == 1)

    mw -= mw / 3
    assert(np.allclose(mw[15], mat * 2))
    assert(mw.npix_allocated == 1)

    mw *= 2
    assert(np.allclose(mw[15], mat * 4))
    assert(mw.npix_allocated == 1)

    mw /= 4
    assert(np.allclose(mw[15], mat))

    mw /= 2
    assert(np.allclose(mw[15], mat / 2))

    m[15] = 10
    mt *= m
    if pol:
        mq *= m
        mu *= m
    mw *= m
    assert(np.allclose(mw[15], mat * 5))
    assert(mw.npix_allocated == 1)

    # weights handling
    if pol:
        remove_weights(mt, mq, mu, mw)
    else:
        remove_weights_t(mt, mw)
    assert(not mt.weighted)
    assert(np.allclose(mw[15], mat * 5))
    ivec = np.linalg.solve(mat * 5, vec * 10) if pol else vec * 2 / mat
    if pol:
        assert(np.allclose([mt[15], mq[15], mu[15]], ivec))
    else:
        assert(np.allclose(mt[15], ivec))

    if pol:
        idet = np.linalg.det(mat * 5)
        det = mw.det()
        assert(np.allclose(det[15], idet))

        icond = np.linalg.cond(mat * 5)
        cond = mw.cond()
        assert(np.allclose(cond[15], icond))

        iinv = np.linalg.inv(mat * 5)
        inv = mw.inv()
        assert(np.allclose(inv[15], iinv))

    if pol:
        assert(np.isnan([mt[16], mq[16], mu[16]]).all())
    else:
        assert(np.isnan(mt[16]))
    assert(not mt.sparse)
    assert(mt.npix_allocated == mt.size)

    if pol:
        apply_weights(mt, mq, mu, mw)
    else:
        apply_weights_t(mt, mw)
    assert(mt.weighted)
    if pol:
        assert(np.allclose([mt[15], mq[15], mu[15]], vec * 10))
        assert(np.isnan([mt[16], mq[16], mu[16]]).all())
    else:
        assert(np.allclose(mt[15], vec * 10))
        assert(np.isnan(mt[16]))
    assert(not mt.sparse)
    assert(mt.npix_allocated == mt.size)

    # compactify maps back to sparse
    mt.compact(zero_nans=True)
    assert(mt.npix_allocated == 1)
    assert(np.allclose(mt[15], np.atleast_1d(vec * 10)[0]))
    mask = mt.to_mask()
    assert(mask[15] == 1)
    assert(mask.sum() == 1)
    assert(mask.to_map().npix_allocated == 1)

    if pol:
        mq.compact(zero_nans=True)
        mu.compact(zero_nans=True)

    # check memory-efficient weights functions
    if pol:
        remove_weights(mt, mq, mu, mw, zero_nans=True)
    else:
        remove_weights_t(mt, mw, zero_nans=True)
    assert(mt.npix_allocated == 1)

    if pol:
        apply_weights(mt, mq, mu, mw)
        assert(np.allclose([mt[15], mq[15], mu[15]], vec * 10))
    else:
        apply_weights_t(mt, mw)
        assert(np.allclose(mt[15], vec * 10))
    assert(mt.npix_allocated == 1)

    if pol:
        # check flatten_pol
        mq2 = mq.copy()
        mu2 = mu.copy()
        mw2 = mw.copy()
        assert(not mq2.flat_pol)
        assert(not mu2.flat_pol)
        flatten_pol(mq2, mu2, mw2)
        assert(mq2.flat_pol)
        assert(not np.allclose(mq2, mq))
        assert(not np.allclose(mu2, mu))
        assert(not np.allclose(mw2.TQ, mw.TQ))
        assert(not np.allclose(mw2.TU, mw.TU))
        assert(not np.allclose(mw2.QQ, mw.QQ))
        assert(not np.allclose(mw2.QU, mw.QU))
        assert(not np.allclose(mw2.UU, mw.UU))

        flatten_pol(mq2, mu2, mw2, invert=True)
        assert(not mq2.flat_pol)
        assert(np.allclose(mq2, mq))
        assert(np.allclose(mu2, mu))
        assert(np.allclose(mw2.TQ, mw.TQ))
        assert(np.allclose(mw2.TU, mw.TU))
        assert(np.allclose(mw2.QQ, mw.QQ))
        assert(np.allclose(mw2.QU, mw.QU))
        assert(np.allclose(mw2.UU, mw.UU))



#!/usr/bin/env python

import spt3g, os
from spt3g import core

args = {"n": 10, "type": core.G3FrameType.Timepoint}
arg = core.G3VectorDouble([1, 2, 3])

p = core.G3Pipeline()
p.Add(core.G3InfiniteSource, **args)
def twiddle(fr, arg=None):
	if fr.type == core.G3FrameType.Timepoint:
		fr['arg'] = arg
p.Add(twiddle, arg=arg)
p.Add(core.Dump)
p.Add(core.G3Writer, filename='testpi.g3')
p.Run()

# Check that pipelineinfo is added and is runnable
assert(len(list(core.G3File('testpi.g3'))) == 11)
for i in core.G3File('testpi.g3'):
	if i.type == core.G3FrameType.PipelineInfo:
		pi = list(i.values())[0]
		break

os.remove('testpi.g3')
print(repr(pi))
pi.Run()

# Check that module arguments survive round-trip to/from storage
mod_args = dict(pi.modules[0])
print(mod_args)
assert(mod_args == args)
obj_arg = pi.modules[1]['arg']
print(obj_arg)
assert(isinstance(obj_arg, type(arg)))
assert((obj_arg == arg).all())

assert(len(list(core.G3File('testpi.g3'))) == 11)

# Check that PI frame has two entries on the second run through
p = core.G3Pipeline()
p.Add(core.G3Reader, filename='testpi.g3')
def check(fr):
	if fr.type == core.G3FrameType.PipelineInfo:
		assert(len(fr) == 2)
p.Add(check)
p.Run()

# Check that PI frame has two entries on the second run through even if a header
# is inserted
print('Test for twiddling')
p = core.G3Pipeline()
p.Add(core.G3Reader, filename='testpi.g3')
def twiddle(fr):
	if fr.type == core.G3FrameType.PipelineInfo:
		return [core.G3Frame(core.G3FrameType.Observation), fr]
p.Add(twiddle)
p.Add(core.G3Writer, filename='testpi2.g3')
p.Run()

# Check that it has three entries when the headered version is run from a file
print('Test for file reading after twiddling')
p = core.G3Pipeline()
p.Add(core.G3Reader, filename='testpi2.g3')
p.Add(core.Dump)
def check(fr):
	if fr.type == core.G3FrameType.PipelineInfo:
		assert(len(fr) == 3)
p.Add(check)
p.Run()

# Check that only one PipelineInfo frame is present, even if it occurs in
# multiple input files.
print('Dedup test 1')
p = core.G3Pipeline()
p.Add(core.G3Reader, filename=['testpi2.g3', 'testpi2.g3'])
p.Add(core.Dump)
i = 0
def check(fr):
	global i
	if fr.type == core.G3FrameType.PipelineInfo:
		assert(len(fr) == 3)
		i += 1
p.Add(check)
p.Run()
assert(i == 1)

# Check that deduplication does not actually deduplicate if there are
# *different* pipeline info frames in multiple input files.
print('Dedup test 2')
p = core.G3Pipeline()
p.Add(core.G3Reader, filename=['testpi.g3', 'testpi2.g3'])
p.Add(core.Dump)
i = 0
def check(fr):
	global i
	if fr.type == core.G3FrameType.PipelineInfo:
		i += 1
p.Add(check)
p.Run()
assert(i == 2)


os.remove('testpi.g3')
os.remove('testpi2.g3')



__version__ = version = '1.0.0'
__version_tuple__ = version_tuple = (1, 0, 0)

fullversion = version
if len(version_tuple) == 2:
    versionname = version
else:
    versionname = 'UNKNOWN'

if len(version_tuple) == 4:
    gitrevision = revision = version_tuple[-1].split('.')[0][1:]
    localdiffs = len(version_tuple[-1].split(".")) == 2
else:
    gitrevision = revision = 'UNKNOWN'
    localdiffs = False

upstream_branch = 'UNKNOWN VCS'
upstream_url = 'UNKNOWN VCS'



#!/usr/bin/env python

from spt3g import core, dfmux
import socket, argparse

parser = argparse.ArgumentParser(description='Record dfmux data to disk')
parser.add_argument('hardware_map', metavar='/path/to/hwm.yaml', help='Path to hardware map YAML file')
parser.add_argument('output', metavar='output', help='Base for output file names')

parser.add_argument('-v', dest='verbose', action='store_true', help='Verbose mode (print all frames)')
parser.add_argument('-a', dest='align', action='store_true', help='Align sampling. This has to happen once, but will break any existing DAN loops when run.')
parser.add_argument('-s', dest='system_time', action='store_true', help='Replace board time with system time when data received. Useful if your board is in IRIG_TEST mode.')
parser.add_argument('--max_file_size', dest='max_file_size', help='Maximum file size in MB')
args = parser.parse_args()

# Import pydfmux later since it can take a while
import pydfmux

print('Initializing hardware map and boards')
hwm = pydfmux.load_session(open(args.hardware_map, 'r'))['hardware_map']
if hwm.query(pydfmux.core.dfmux.IceCrate).count() > 0:
        hwm.query(pydfmux.core.dfmux.IceCrate).resolve()

if args.align:
	print('Aligning board sampling, this will break any existing DAN loops!')
	hwm.query(pydfmux.core.dfmux.IceBoard).set_fir_stage(6)
	hwm.query(pydfmux.core.dfmux.IceBoard).align_sampling()

print('Beginning data acquisition')
# Set up DfMux consumer
pipe = core.G3Pipeline()
builder = dfmux.DfMuxBuilder([int(board.serial) for board in hwm.query(pydfmux.IceBoard)])

# Get the local IP to connect to the boards by opening a test connection.
testsock = socket.create_connection(('iceboard' + str(hwm.query(pydfmux.core.dfmux.IceBoard).first().serial) + '.local', 80))
local_ip = testsock.getsockname()[0]
testsock.close()

collector = dfmux.DfMuxCollector(local_ip, builder)
pipe.Add(builder)

# Insert current hardware map into data stream. This is critical to get the
# board ID -> IP mapping needed to do anything useful with the data
pipe.Add(dfmux.PyDfMuxHardwareMapInjector, pydfmux_hwm=hwm)

if args.system_time:
	def sub_system_time(frame):
		if frame.type != core.G3FrameType.Timepoint:
			return
		del frame['EventHeader']
		frame['EventHeader'] = core.G3Time.Now()
	pipe.Add(sub_system_time)

pipe.Add(dfmux.PeriodicHousekeepingCollector)
pipe.Add(dfmux.HousekeepingConsumer, subprocess=True)

if args.verbose:
	pipe.Add(core.Dump)

pipe.Add(core.G3Writer, filename=args.output)

collector.Start()
pipe.Run()




#!/usr/bin/env python

from spt3g import core

frames = []
frames.append(core.G3Frame(core.G3FrameType.Calibration))
frames.append(core.G3Frame(core.G3FrameType.Calibration))
frames.append(core.G3Frame(core.G3FrameType.Scan))
frames.append(core.G3Frame(core.G3FrameType.Calibration))
frames.append(core.G3Frame(core.G3FrameType.Calibration))
frames.append(core.G3Frame(core.G3FrameType.Scan))
frames.append(core.G3Frame(core.G3FrameType.Scan))

# Make first two cal frames different, the rest identical
frames[0]['Seq'] = 'B'
frames[1]['Seq'] = 'A'
frames[3]['Seq'] = 'A'
frames[4]['Seq'] = 'A'

# Make all Scan frames identical
frames[2]['Seq'] = 'B'
frames[5]['Seq'] = 'B'
frames[6]['Seq'] = 'B'

# Result should be that we get the first two calibration frames, since they are
# different, all the scan frames, since it should not even try to deduplicate
# data, and nothing else

out = []
pipe = core.G3Pipeline()

sent = False
def framesource(fr):
	global sent
	if not sent:
		sent = True
		return frames
	else:
		return []
pipe.Add(framesource)

pipe.Add(core.DeduplicateMetadata)

def getframes(fr):
	if fr.type != core.G3FrameType.EndProcessing and fr.type != core.G3FrameType.PipelineInfo:
		out.append(fr)
pipe.Add(getframes)

pipe.Run()

assert(len(out) == 5)
assert(out[0].type == core.G3FrameType.Calibration)
assert(out[0]['Seq'] == 'B')
assert(out[1].type == core.G3FrameType.Calibration)
assert(out[1]['Seq'] == 'A')


assert(out[2].type == core.G3FrameType.Scan)
assert(out[3].type == core.G3FrameType.Scan)
assert(out[4].type == core.G3FrameType.Scan)




import spt3g._libcalibration
for k in dir(spt3g._libcalibration):
    print(k, getattr(spt3g._libcalibration, k))
from spt3g._libcalibration import BolometerProperties
import spt3g.calibration
from spt3g import calibration
from spt3g.calibration import BolometerProperties



from .. import core
from . import BolometerProperties, BolometerPropertiesMap
from . import PointingProperties, PointingPropertiesMap
import numpy, os, re

'''
Utilities for bringing together calibration data from a number of sources
to produce unified calibration data to go in folders with the data for
analysis processing.
'''

@core.indexmod
class BuildBoloPropertiesMap(object):
    '''
    Build bolometer properties map from raw calibration sub-frames emitted
    by other processing scripts. Input data are maps of bolo ID to floats
    or, in the case of names, strings. If multiple instance of each data
    class appear, the final bolometer properties map will be the median
    of the input values.

    Expects to be passed frames from:
    - RCW38 Relative Pointing Offset Calibration (Keys: 'PointingOffsetX', 'PointingOffsetY')
    - CenA Angle Calibration (Keys: 'PolarizationAngle', 'PolarizationEfficiency')
    - Band Calibration (Key: 'BoloBands')
    - Physical Name Data (Key: 'PhysicalBoloIDs')
    '''
    
    def __init__(self, drop_original_frames=True, fiducial_detectors=[],
                 bpm_name='NominalBolometerProperties', use_bpm_pointing=False):
        '''
        If drop_original_frames is True, will drop all input Calibration frames.

        If fiducial_detectors is set, will use the average of the position[s] of 
        whatever detector[s] are specified to center each set of relative offsets
        encountered (NB: this recentering is done in a Cartesian way). If it is
        *not* specified, five detectors near the middle of the focal plane present
        in every observation given will be chosen automatically and printed to the
        console.

        bpm_name is the name of the key containing the input BolometerPropertiesMap
        that will be merged into the output map.

        If use_bpm_pointing is True, then the pointing information is extracted
        from the input BolometerPropertiesMap.  If False, pointing information
        must be supplied in an additional input frame.
        '''

        self.drop_original_frames = drop_original_frames
        self.fiducial_detectors = fiducial_detectors
        self.bpm_name = bpm_name
        self.use_bpm_pointing = use_bpm_pointing

        self.props = {}
        self.dfmux_tf = None

    def __call__(self, frame):
        if frame.type == core.G3FrameType.EndProcessing:
            boloprops = BolometerPropertiesMap()

            # Technique to average together points while ignoring outliers
            def robust_avg(data):
                from scipy.stats import sigmaclip

                gooddata = numpy.asarray(data)[numpy.isfinite(data)]
                if len(gooddata) == 1:
                    return gooddata[0]
                return numpy.median(sigmaclip(gooddata, low=2.0, high=2.0)[0])

            if len(self.fiducial_detectors) == 0:
                always_dets = set([bolo for bolo in self.props if numpy.isfinite(self.props[bolo].get('xoffsets', [numpy.nan])).all()])
                self.fiducial_detectors = sorted(always_dets, key=lambda bolo: robust_avg(self.props[bolo]['xoffsets'])**2 + robust_avg(self.props[bolo]['yoffsets'])**2)[:5]
                print('Using %s as the set of fiducial detectors for pointing' % self.fiducial_detectors)

            for bolo in self.props.keys():
                p = BolometerProperties()

                if 'xoffsets' in self.props[bolo]:
                    # Ideally we would have error bars on these measurements...
                    p.x_offset = robust_avg([self.props[bolo]['xoffsets'][i] - numpy.mean([self.props[fd]['xoffsets'][i] for fd in self.fiducial_detectors]) for i in range(len(self.props[bolo]['xoffsets']))])
                    p.y_offset = robust_avg([self.props[bolo]['yoffsets'][i] - numpy.mean([self.props[fd]['yoffsets'][i] for fd in self.fiducial_detectors]) for i in range(len(self.props[bolo]['xoffsets']))])

                if 'physname' in self.props[bolo]:
                    p.physical_name = self.props[bolo]['physname']

                if 'band' in self.props[bolo]:
                    p.band = self.props[bolo]['band']

                if 'polangle' in self.props[bolo]:
                    # Ideally we would have error bars on these measurements.

                    # Strategy: to avoid wraparound issues, form a 2-D vector
                    # and then take the median X/Y position
                    # XXX: Do anything about vectors being headless?
                    angles = numpy.asarray(self.props[bolo]['polangle'])
                    efficiencies = numpy.asarray(self.props[bolo]['poleff'])
                    poly = robust_avg(numpy.sin(angles/core.G3Units.rad)*efficiencies)
                    polx = robust_avg(numpy.cos(angles/core.G3Units.rad)*efficiencies)
                    p.pol_angle = numpy.arctan2(poly, polx)
                    p.pol_efficiency = numpy.hypot(poly, polx)

                if 'coupling' in self.props[bolo]:
                    p.coupling = self.props[bolo]['coupling']

                if 'pixel_id' in self.props[bolo]:
                    p.pixel_id = self.props[bolo]['pixel_id']
                if 'wafer_id' in self.props[bolo]:
                    p.wafer_id = self.props[bolo]['wafer_id']
                if 'pixel_type' in self.props[bolo]:
                    p.pixel_type = self.props[bolo]['pixel_type']
                
                boloprops[bolo] = p

            cframe = core.G3Frame(core.G3FrameType.Calibration)
            cframe['BolometerProperties'] = boloprops
            cframe['FiducialPointingDetectors'] = core.G3VectorString(self.fiducial_detectors)
            cframe['DfMuxTransferFunction'] = self.dfmux_tf
            return [cframe, frame]

        if self.bpm_name in frame:
            bpm = frame[self.bpm_name]
            if len(self.props) > 0:
                npointings = max([len(p.get('xoffsets', [])) for p in self.props.values()])
            else:
                npointings = 0

            for bolo in bpm.keys():
                bp = bpm[bolo]

                if bolo not in self.props:
                    self.props[bolo] = {}

                if self.use_bpm_pointing:
                    if 'xoffsets' not in self.props[bolo]:
                        self.props[bolo]['xoffsets'] = [numpy.nan]*npointings
                        self.props[bolo]['yoffsets'] = [numpy.nan]*npointings

                    self.props[bolo]['xoffsets'].append(bp.x_offset)
                    self.props[bolo]['yoffsets'].append(bp.y_offset)

                self.props[bolo]['band'] = bp.band
                self.props[bolo]['coupling'] = bp.coupling
                self.props[bolo]['physname'] = bp.physical_name
                self.props[bolo]['pixel_id'] = bp.pixel_id
                self.props[bolo]['wafer_id'] = bp.wafer_id
                self.props[bolo]['pixel_type'] = bp.pixel_type

                if 'polangle' not in self.props[bolo]:
                    self.props[bolo]['polangle'] = []
                    self.props[bolo]['poleff'] = []
                self.props[bolo]['polangle'].append(bp.pol_angle)
                self.props[bolo]['poleff'].append(bp.pol_efficiency)

            # Book NaNs for any non-present detectors
            if self.use_bpm_pointing:
                for bolo,p in self.props.items():
                    if bolo in bpm:
                        continue
                    if 'xoffsets' in p:
                        self.props[bolo]['xoffsets'].append(numpy.nan)
                        self.props[bolo]['yoffsets'].append(numpy.nan)

        # Pointing calibration
        if 'PointingOffsetX' in frame:
            if len(self.props) > 0:
                npointings = max([len(p.get('xoffsets', [])) for p in self.props.values()])
            else:
                npointings = 0
            for bolo in frame['PointingOffsetX'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                if 'xoffsets' not in self.props[bolo]:
                    self.props[bolo]['xoffsets'] = [numpy.nan]*npointings
                    self.props[bolo]['yoffsets'] = [numpy.nan]*npointings

                self.props[bolo]['xoffsets'].append(frame['PointingOffsetX'][bolo])
                self.props[bolo]['yoffsets'].append(frame['PointingOffsetY'][bolo])
            for bolo,p in self.props.items():
                if bolo in frame['PointingOffsetX']:
                    continue
                if 'xoffsets' in p:
                    self.props[bolo]['xoffsets'].append(numpy.nan)
                    self.props[bolo]['yoffsets'].append(numpy.nan)

        # Band calibration
        if 'BoloBands' in frame:
            for bolo in frame['BoloBands'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                #assert('band' not in self.props[bolo] or self.props[bolo]['band'] == frame['BoloBands'][bolo])
                self.props[bolo]['band'] = frame['BoloBands'][bolo]

        # Optical coupling
        if 'BoloCoupling' in frame:
            for bolo in frame['BoloCoupling'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                self.props[bolo]['coupling'] = frames['BoloCoupling'][bolo]

        # Names
        if 'PhysicalBoloIDs' in frame:
            for bolo in frame['PhysicalBoloIDs'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                #assert('physname' not in self.props[bolo] or self.props[bolo]['physname'] == frame['PhysicalBoloIDs'][bolo])
                self.props[bolo]['physname'] = frame['PhysicalBoloIDs'][bolo]

        if 'PixelIDs' in frame:
            for bolo in frame['PixelIDs'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                #assert('pixel_id' not in self.props[bolo] or self.props[bolo]['pixel_id'] == frame['PixelIDs'][bolo])
                self.props[bolo]['pixel_id'] = frame['PixelIDs'][bolo]

        if 'WaferIDs' in frame:
            for bolo in frame['WaferIDs'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                #assert('wafer_id' not in self.props[bolo] or self.props[bolo]['wafer_id'] == frame['WaferIDs'][bolo])
                self.props[bolo]['wafer_id'] = frame['WaferIDs'][bolo]

        if 'PixelTypes' in frame:
            for bolo in frame['PixelTypes'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                # assert('pixel_type' not in self.props[bolo] or self.props[bolo]['pixel_type'] == frame['PixelTypes'][bolo])
                self.props[bolo]['pixel_type'] = frame['PixelTypes'][bolo]

        # Polarization Angles
        if 'PolarizationAngle' in frame:
            for bolo in frame['PolarizationAngle'].keys():
                if bolo not in self.props:
                    self.props[bolo] = {}
                if 'polangle' not in self.props[bolo]:
                    self.props[bolo]['polangle'] = []
                    self.props[bolo]['poleff'] = []

                self.props[bolo]['polangle'].append(frame['PolarizationAngle'][bolo])
                self.props[bolo]['poleff'].append(frame['PolarizationEfficiency'][bolo])

        # DfMux Transfer Function
        if 'DfMuxTransferFunction' in frame:
            self.dfmux_tf = frame['DfMuxTransferFunction']

        if frame.type == core.G3FrameType.Calibration and self.drop_original_frames:
            return []

@core.indexmod
def ExplodeBolometerProperties(frame, bpmname='NominalBolometerProperties'):
    '''
    Take a bolometer properties map (usually the nominal one) and convert it
    into its constituent keys as though they came from real calibration. This
    is the inverse of BuildBoloPropertiesMap and mostly is useful when combining
    hardware map information with real calibration.
    '''

    if bpmname not in frame:
        return

    bpm = frame[bpmname]
    polangle = core.G3MapDouble()
    poleff = core.G3MapDouble()
    bands = core.G3MapDouble()
    names = core.G3MapString()
    pixels = core.G3MapString()
    wafers = core.G3MapString()
    pixtypes = core.G3MapString()
    xoff = core.G3MapDouble()
    yoff = core.G3MapDouble()
    coupling = core.G3MapInt()

    for bolo, p in bpm.items():
        polangle[bolo] = p.pol_angle
        poleff[bolo] = p.pol_efficiency
        bands[bolo] = p.band
        names[bolo] = p.physical_name
        xoff[bolo] = p.x_offset
        yoff[bolo] = p.y_offset
        pixels[bolo] = p.pixel_id
        wafers[bolo] = p.wafer_id
        pixtypes[bolo] = p.pixel_type
        coupling[bolo] = p.coupling

    frame['PolarizationAngle'] = polangle
    frame['PolarizationEfficiency'] = poleff
    frame['PhysicalBoloIDs'] = names
    frame['BoloBands'] = bands
    frame['PointingOffsetX'] = xoff
    frame['PointingOffsetY'] = yoff
    frame['PixelIDs'] = pixels
    frame['WaferIDs'] = wafers
    frame['PixelTypes'] = pixtypes
    frame['BoloCoupling'] = coupling

@core.indexmod
class BuildPointingProperties(object):
    '''
    Build pointing properties from raw calibration sub-frames emitted
    by other processing scripts. Input data are floats, including
    tilt information, and eventually other pointing model values.

    Expects to be passed frames from:
    - Az tilt fit parameters (Keys: 'tiltAngle', 'tiltHA', 'tiltLat', 'tiltMag')
    '''
    
    def __init__(self, drop_original_frames=True):
        '''
        If drop_original_frames is True, will drop all input Calibration frames.
        '''

        self.drop_original_frames = drop_original_frames

        self.props = {}

    def __call__(self, frame):
        if frame.type == core.G3FrameType.EndProcessing:
            pointingprops = PointingPropertiesMap()

            p = PointingProperties()

            if 'tiltLat' in self.props:
                p['tiltLat'] = self.props['tiltLat']

            if 'tiltHA' in self.props:
                p['tiltHA'] = self.props['tiltHA']
            
            if 'tiltMag' in self.props:
                p['tiltMag'] = self.props['tiltMag']
            
            if 'tiltAngle' in self.props:
                p['tiltAngle'] = self.props['tiltAngle']

        if frame.type == core.G3FrameType.Calibration and self.drop_original_frames:
            return []

@core.indexmod
class MergeCalibrationFrames(object):
    '''
    Merge the keys from a sequence of calibration frames. Will throw an
    exception if a key recurs in more than one calibration frame. The merged
    calibration frame will be emitted before the first non-calibration frame
    that follows a calibration frame or at the end of processing, whichever
    comes first. Other non-calibration frames will be ignored.
    '''
    def __init__(self, KeysToIgnore=['PointingOffsetX', 'PointingOffsetY']):
        '''
        Ignores keys in the KeysToIgnore list during merging. By default, set
        to values written by the flux/pointing calibration that are stored to
        the BolometerPropertiesMap.
        '''
        self.outframe = core.G3Frame(core.G3FrameType.Calibration)
        self.ignore_keys = KeysToIgnore
    def __call__(self, frame):
        if frame.type == core.G3FrameType.Calibration:
            for k in frame.keys():
                if k not in self.ignore_keys:
                    self.outframe[k] = frame[k]
            return []

        # Allow extra calibration frames to follow proper calframe files
        # (which likely have PipelineInfo frames in them)
        if frame.type == core.G3FrameType.PipelineInfo:
            return

        # Ignore random wiring frames etc.
        if self.outframe is None or len(self.outframe.keys()) == 0:
            return

        # Return merged frame before whatever follows if we have new data
        out = [self.outframe, frame]
        self.outframe = None
        return out




# We need to share certain global variables (cereal and pybind11
# type registries between shared objects, so make sure the RTLD
# is configured to allow that to happen. This can sometimes work
# without this on glibc due to a glibc bug, but it will fail on
# other libc implementations
import os, sys
sys.setdlopenflags(sys.getdlopenflags() | os.RTLD_GLOBAL)

from .._libcore import *

from .modconstruct import usefulfunc, pipesegment, indexmod
from .g3logging import *

from .fileio import *
try:
	from .parse_pipeline_graph import plot_frame_processing_info
except ImportError:
	pass
try:
	from .multiprocess import Subproc
except ImportError:
	pass
from .util import *

from .apidoc import *
from .timestreamextensions import *
from .quatextensions import *

from .g3decorators import cache_frame_data, scan_func_cache_data



#!/usr/bin/env python

import numpy
from spt3g import core

ts = core.G3Timestream(numpy.zeros(50))
ts.units = core.G3TimestreamUnits.Power
ts.start = core.G3Time(0)
ts.stop = core.G3Time(10*core.G3Units.s)

assert(ts[5] == 0) # Check scalar getitem

ts[12] = 13.4 # Check scalar setitem
assert(ts[12] == 13.4)

ts[:] = numpy.arange(50) # Test vector setitem
assert(ts[12]) == 12.0

# Test units preserved by slicing
assert(ts[:5].units == ts.units)

# Test length with slicing
assert(len(ts[::2]) == len(ts)/2)
assert(len(ts[5:]) == len(ts)-5)

# Test consistency with numpy slicing with steps that do and do not divide evenly into array length
assert((numpy.asarray(ts[::2]) == numpy.asarray(ts)[::2]).all())
assert((numpy.asarray(ts[::3]) == numpy.asarray(ts)[::3]).all())
assert((numpy.asarray(ts[5::2]) == numpy.asarray(ts)[5::2]).all())
assert((numpy.asarray(ts[5::3]) == numpy.asarray(ts)[5::3]).all())
assert((numpy.asarray(ts[5:-4:2]) == numpy.asarray(ts)[5:-4:2]).all())
assert((numpy.asarray(ts[:-4:2]) == numpy.asarray(ts)[:-4:2]).all())
assert((numpy.asarray(ts[:-4]) == numpy.asarray(ts)[:-4]).all())
assert((numpy.asarray(ts[5:]) == numpy.asarray(ts)[5:]).all())
assert((numpy.asarray(ts[5:-4]) == numpy.asarray(ts)[5:-4]).all())

# Test sample rate consistency for the first and last N samples
assert(numpy.isclose(ts[:5].sample_rate, ts.sample_rate, atol=0))
assert(numpy.isclose(ts[-5:].sample_rate, ts.sample_rate, atol=0))

# Test start and stop time accuracy
assert(numpy.allclose([t.time for t in ts[:5].times], [t.time for t in ts.times][:5]))
assert(numpy.allclose([t.time for t in ts[:5].times], [t.time for t in ts.times[:5]]))
assert(numpy.allclose([t.time for t in ts[-10:].times], [t.time for t in ts.times][-10:]))
assert(numpy.allclose([t.time for t in ts[-10:].times], [t.time for t in ts.times[-10:]]))
assert(numpy.allclose([t.time for t in ts[5:-10].times], [t.time for t in ts.times][5:-10]))
assert(numpy.allclose([t.time for t in ts[5:-10].times], [t.time for t in ts.times[5:-10]]))

# Test that we get the right sample rates when sparsely sampling
assert(numpy.isclose(ts[::5].sample_rate, ts.sample_rate/5, atol=0))
assert(numpy.isclose(ts[::2].sample_rate, ts.sample_rate/2, atol=0))

# Mix all the slicing tests together in a big bucket
assert(numpy.isclose(ts[12::2].sample_rate, ts.sample_rate/2, atol=0))
assert(numpy.isclose(ts[12:-2:2].sample_rate, ts.sample_rate/2, atol=0))
assert(numpy.isclose(ts[:-2:2].sample_rate, ts.sample_rate/2, atol=0))




from spt3g import core, maps
import numpy as np

maplist = [
    maps.FlatSkyMap(300, 300, core.G3Units.arcmin, proj=maps.MapProjection.ProjZEA),
    maps.HealpixSkyMap(64),
]

for m in maplist:

    pipe = core.G3Pipeline()

    pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Observation, n=1)
    pipe.Add(maps.InjectMapStub, map_id="test_map", map_stub=m)

    def RandomMap(frame):
        if frame.type != core.G3FrameType.Map:
            return

        tmap = frame.pop("T")
        tmap[:] = np.random.randn(*m.shape)
        frame["T"] = tmap

        wmap = frame.pop("Wunpol")
        wmap.TT[:] = 10 * np.ones(m.shape)
        frame["Wunpol"] = wmap


    pipe.Add(RandomMap)

    mex0 = maps.ExtractMaps(copy=True)
    pipe.Add(mex0)

    pipe.Add(maps.MakeMapsPolarized)

    pipe.Add(maps.ValidateMaps)
    pipe.Add(maps.SetPolConv, pol_conv=maps.MapPolConv.COSMO)
    pipe.Add(maps.ValidateMaps)

    pipe.Add(maps.RemoveWeights)
    pipe.Add(maps.ApplyWeights)

    mex1 = maps.ExtractMaps()
    pipe.Add(mex1)

    pipe.Add(
        maps.ReplicateMaps,
        input_map_id="test_map",
        output_map_ids=["test_map1", "test_map2"],
        copy_weights=True,
    )

    coadder = maps.CoaddMaps(keep_outputs=True)
    pipe.Add(coadder)

    mex2 = maps.ExtractMaps(copy=True)
    pipe.Add(mex2)

    pipe.Add(maps.MakeMapsUnpolarized)

    mex3 = maps.ExtractMaps()
    pipe.Add(mex3)

    tmap = m.array_clone(np.random.randn(*m.shape))
    tmap.pol_type = maps.MapPolType.T
    tmap.weighted = False
    pipe.Add(maps.InjectMaps, map_id="test_map", maps_in=[tmap])

    mex4 = maps.ExtractMaps()
    pipe.Add(mex4)

    pipe.Run()

    # check that mex0 and mex1 are nearly (but not exactly) identical
    assert np.max(mex0.maps["test_map"]["T"] - mex1.maps["test_map"]["T"]) != 0
    assert np.allclose(mex0.maps["test_map"]["T"], mex1.maps["test_map"]["T"])

    # check that replication dropped the test map
    assert "test_map" not in mex2.maps
    assert "test_map" not in mex3.maps

    # check that injection added a new test map
    assert "test_map" in mex4.maps
    assert (mex4.maps["test_map"]["T"] == tmap).all()

    # check that coadd is correct
    assert "Coadd" in mex2.maps
    assert list(coadder.coadd_frame["InputMapIds"]) == ["test_map1", "test_map2"]
    assert (
        mex2.maps["Coadd"]["T"] ==
        mex2.maps["test_map1"]["T"] + mex2.maps["test_map2"]["T"]
    ).all()
    assert "Coadd" in mex3.maps
    assert (
        mex3.maps["Coadd"]["T"] ==
        mex3.maps["test_map1"]["T"] + mex3.maps["test_map2"]["T"]
    ).all()

    # check that replicated maps are properly populated
    for map_id in ["test_map1", "test_map2"]:
        assert map_id in mex2.maps
        assert map_id in mex3.maps

        mdict = mex2.maps[map_id]
        for k in ["T", "Q", "U", "Wpol"]:
            assert k in mdict
            assert mdict[k].compatible(m)

        mdict = mex3.maps[map_id]
        for k in ["T", "Wunpol"]:
            assert k in mdict
            assert mdict[k].compatible(m)



# -*- coding: utf-8 -*-
'''
'''
import unittest
import math


# From PyWTL. In the unlikely event PyWTL is ever updated, please synchronize.

__all__ = ['convert_mb', 'convert_mezz', 'convert_squid', 'convert_demod']

class wtl_convert:
    '''
       Equations for basic conversions to be inheritted by other derives classes.
       All methods in this class are static
       (there are no data members of the class), meaning there is never any reason to instantiate
       an instance of this class. You can always safely use  wtl_convert.method().
    '''
    #_v1 = 3.3
    global _Vadc, _drHK
    _Vadc = 3.3
    _drHK = 2**12

    def _v3_voltage_divider(self, r1, r2, r3, v1, v2):
        v3 = ((r3*v1)/r1) - (((r1*r2+r2*r3+r1*r3)*v2)/(r1*r2))
        return -v3

    def _v1_voltage_divider(self, r1, r2, v2):
        return v2 * (1 + (r1/r2))

    def _count_to_voltage(self, adc_count):
        return adc_count * (_Vadc/_drHK)

    def _voltage_to_count(self, voltage):
        return voltage * (_drHK/_Vadc)

    def temperature(self, adc_count):
        if(adc_count >= 8192):
            adc_count = adc_count - 16384;
        return adc_count / 32.0

class convert_mezz(wtl_convert):
    '''
        Handles conversions for the mezzanine board. Inherits from wtl_convert.

    '''
    def neg_voltage(self, adc_count_pos, adc_count_neg):
        return self._v3_voltage_divider(10000, 10000, 100000,
                                        self.pos_voltage(adc_count_pos),
                                        self._count_to_voltage(adc_count_neg))

    def pos_voltage(self, adc_count):
        return self._v1_voltage_divider(20000, 10000, self._count_to_voltage(adc_count))

    def demod_voltage_se(self, adc_count_pos, adc_count_demod):
        return self._v3_voltage_divider(10000, 5000, 20050,
                                        self.pos_voltage(adc_count_pos),
                                        self._count_to_voltage(adc_count_demod))

    def demod_voltage_de(self, adc_count_pos, adc_count_demod_pos, adc_count_demod_neg):
        vp = self._v3_voltage_divider(10000, 10000, 5050,
                                      self.pos_voltage(adc_count_pos),
                                      self._count_to_voltage(adc_count_demod_pos))
        vn = self._v3_voltage_divider(10000, 10000, 5050,
                                      self.pos_voltage(adc_count_pos),
                                      self._count_to_voltage(adc_count_demod_neg))
        return vp-vn




class convert_mb(wtl_convert):
    '''
    '''
    global _v1
    _v1 = 3.3

    def __init__(self, units='Normalized_16ch'):
        if units=='Normalized':
            '''
            This converts a 24 bit number to Normalized units. Useful for e.g.
            GCP housekeeping data.
            TdH - 19Apr2012
            '''
            self.human2amp_fact = 2.**23 - 1.
        elif units=='Normalized_16ch':
            # I'm not sure what this is, but I am leaving it in place.
            self.human2amp_fact = 2.**16 - 1.
        else:
            raise ValueError('convert_mb :: Unknown argument %s' % str(units))

    def neg_voltage(self, adc_count):
        return self._v3_voltage_divider(10000, 10000, 100000,
                                        _v1,
                                        self._count_to_voltage(adc_count))

    def pos_voltage(self, adc_count):
        return self._v1_voltage_divider(30100, 10000, self._count_to_voltage(adc_count))

    def freq_to_human(self, raw_frequency):
        return (raw_frequency / (2.**32-1)) * 25.e6

    def human_to_freq(self, f,round=3):
        return self.round_and_convert_frequencyHz_to_raw(f,round)

    def phase_to_human(self, raw_phase, rads=False):
        if rads:
            return (raw_phase / (2.**32-1)) * 2 * math.pi
        return (raw_phase / (2.**32-1)) * 360.

    def phase_to_raw(self, human_phase, rads=False):
        if rads:
            return human_phase*(2.**32-1)/(2*math.pi)
        else:
            human_phase%360 #Make phase between 0 and 360.
            return (human_phase*2.**32-1)/360.

    def amp_to_human(self, raw_amplitude):
        return raw_amplitude / self.human2amp_fact

    def human_to_amp(self, human_amplitude):
        return int(human_amplitude * self.human2amp_fact)

        def amp_to_raw(self, human_amp):
                return human_amp * (2.**16-1)


    def round_and_convert_frequencyHz_to_raw( self, freqHz, round=3 ):
        '''
            helper routine for set_frequency
            Funny things happen when you ask a digital sine wave generator to operate at integer multiples of the clock.
            Our clock is 25 MHz. To ensure this never happens, we ONLY allow you to set frequencies that are
            multiples of 3 Hz.
            THIS FUNCTION IS NOT SUFFICIENT. Still have a problem if the difference from a fraction of the
            clock is, itself, a fraction of the clock as the samples repeat each nth time and form noise
            spikes. New function pending.
        '''
        intFreqHz = integer_round_off(  ( integer_round_off(freqHz/round) ) * round  )
        freqRaw = integer_round_off( intFreqHz / 25000000.0 * (2**32 - 1) )
        return freqRaw

    def freqHz_to_raw(self,freq,rounding=3):
        rounded = (int(freq)/rounding)*rounding
# Work around for python bug - casting floats of a certain magnitude to integers -> 0
# Raw frequencies starting at ~6.2MHz (raw = 2**30) set it off.
# Float to long and long to int is safe.
        raw = int(int(round((rounded/25e6)*(2**32-1))))
        return raw


class convert_squid:
    '''
        Class for conversion from settings to physical units at the SQUID.
    '''
    global _dr, _Vdac
    _dr = 2**14
    _Vdac = 3.0

    def __init__(self, units=None):
      if units==None:
        print("")
        print("convert_squid was not called with the 'units' argument.")
        print("Please specify whether you're using dfmux.NORMALIZED or")
        print("dfmux.NORMALIZED_16CH units. e.g.")
        print("  convert_squid(units=b.NORMALIZED)")
        print("For those running the old (classic / pre-DAN) firmware")
        print("I'll assume you're using NORMALIZED_16CH for now,")
        print("but please replace")
        print("  convert_squid()")
        print("by")
        print("  convert_squid('Normalized_16ch')")
        print("in your code and don't worry about this.")
        print("")
        self.units = 'Normalized_16ch'
      else:
        self.units = units
      if self.units=='Normalized':
        self.amplitude_to_normalized16ch_conversion = 16.
      elif self.units=='Normalized_16ch':
        self.amplitude_to_normalized16ch_conversion = 1.
      else:
        raise Exception("convert_squid was called with an erroneous 'units' argument. Please provide dfmux.NORMALIZED or dfmux.NORMALIZED_16CH")

    def toVoltageSquidDACOffset(self, value):
        '''
            Converts raw output of squid control board DAC Offset field to volts at the DAC. This
            field sets the zero point of the DAC.
        '''
        return 6 * (value * _Vdac) / _dr

    def toVoltageSquidChannel(self, value):
        '''
            Converts raw output of squid control board channel to volts at the DAC. This applies
            to all squid board channels. ie: squid biases, flux biases, heaters and stage1_offset.
        '''
        return ((7 * _Vdac * value) / _dr) - 8

    def toRawVoltage(self, voltage):
        '''
            Converts voltage at DAC to raw (machine number) for setting on the squid control board
        '''
        return math.floor((((voltage+8.0)/7.0) * _dr)/_Vdac)

    def toVsquid(self, value, type='default'):
        ''' converts Voffset voltage to voltage at input of the 1st stage
            amplifier.  This is also the output voltage of the Squid

            type is made for custom changes.
                default : typical DfMUX and SQUID controller board
                EBEX : Rx36 and Rx43 are changed from 5K to 2.5K in EBEX, the
                       transfer function is 20/15K instead of 20/20K
        '''
        r1 = 20.
        r2 = 20.06e3
        if type=='EBEX':
            value = value*4./3.
        return value*r1/(r1+r2)

    def toFluxQuanta(self, value):
        ''' converts DAC units to DC flux applied to the SQUID
            in units of the flux quanta
        '''
        R = 50.28e3
        M = 26e-6     # mutual inductance of the squid 26uA/phi_o
        return value/R/M     # flux to dac units

    def fluxQuantaToVFluxBias(self, value):
        '''
            converts from fraction of flux quanta to the voltage difference
            at the flux bias coil. (Inverse of toFluxQuanta)
        '''
        R = 50.28e3
        M = 26e-6
        return value*R*M

    def toFluxCurrent(self, value):
        '''
        Converts V_flux_bias at the DAC to current at the flux bias inductor in
        ** Amps **
        '''
        R = 50.28e3
        return value/R

    def toSquidCurrent(self, value):
        '''
        Converts V_SQUID_bias at the DAC to current at the SQUID in
        ** Amps **
        '''
        R = 50.28e3
        return value/R

    def nullerAmpToSquidCurrent(self, amp, gain, frequency=500000):
        '''
        Convert a nuller amplitude setting to the current at the SQUID coil. Based
        on measurements done at McGill, see the following wiki page for details:
        http://kingspeak.physics.mcgill.ca/twiki/bin/view/DigitalFMux/T_null

        Notes:
            - Based on 16 channel non-DAN/DAF firmware "Antlion"

        Input units:
        amp - Nuller fractional amplitude
        gain - nuller gain number (0,1,2,3)
        frequency - Hertz

                Output units:
                current - Arms
        '''
        amp = amp * self.amplitude_to_normalized16ch_conversion
        Tgain = [1.50, 3.39, 10.4, 29.7]
        return Tgain[gain]/(1. + (float(frequency)/2.365E6)**3)**(0.5) * amp * 10**(-6)


    def theo_AnToIbolo(self, An, Gnuller, N=16, eff=0.9): # Kam Arnolds's first cut measurement of efficiency`
        '''
        Theoretical convertion from fractionnal nuller amplitude into
            A_rms at the SQUID coil from schematics
            Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        An is the nuller fractionnal amplitude (either a pylab array or
            a scalar)
        Gnuller is the nuller gain (0-3)
        N is the multiplexing factor (8 or 16)
        returns the A_rms through the SQUID coil
        NOTE: The input (An) is an *amplitude*, the output is an *RMS* (7Nov2011 - TdH)
        '''
        '''
        if not eff==1:
            print '\n##############################################################'
            print '# Warning : An efficiency of '+str(eff)+' is used in theo_AnToIbolo.  #'
            print '# So this function does dot reflect the theoretical transfer #'
            print '# function as it is supposed.                                #'
            print '##############################################################\n'
            '''
        An = An * self.amplitude_to_normalized16ch_conversion

        #variable resistor on demod chain (mezz rev2 and 3)
        R=[2000.0, 820.0, 200.0, 0.0]

        #transfer function
        I = An * 3.3179e-3/(100.0+R[Gnuller])

        #multiplexing factor
        if N==8:
            I*=2.0
        elif N==16:
            pass
        else:
            return -1

        return I*eff


    def theo_IboloToAn(self, I, Gnuller, N=16, eff=0.9):
        '''
        Theoretical convertion from A_rms at the SQUID coil  to
            fractionnal nuller amplitude from schematics
            Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        I is the A_rms through the SQUID coil
        Gnuller is the nuller gain (0-3)
        N is the multiplexing factor (8 or 16)
        returns the nuller fractionnal amplitude
        '''
        #variable resistor on demod chain (mezz rev2 and 3)
        R=[2000.0, 820.0, 200.0, 0.0]

        #transfer function
        An = I * (100.0+R[Gnuller])/3.3179e-3

        #multiplexing factor
        if N==8:
            An/=2.0
        elif N==16:
            pass
        else:
            return -1

        return An / self.amplitude_to_normalized16ch_conversion/eff


    def theo_AcToVbias(self,Ac,Gcarrier,N=16, correct_for_efficiency=False):
        '''
        Theoretical conversion from fractional carrier amplitude to
            V_rms across bias resistor. Calculated using schematics
            Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        Arguments:

          Ac - is the carrier fractional amplitude and should be in the units
               provided to convert_squid. Note this can be a single value, or
               and array.

          Gcarrier - the carrier gain setting (0-3)

          N  - This is a hack which you should ignore unless you are running eight
               channel firmware, in which case you should provide convert_squid
               with 'Normalized_16ch' and set N=8, here.

          correct_for_efficiency - Set this to True if you want some kind of
                                   correction to be applied. I have no idea where
                                   this correction came from so I wouldn't do it
                                   (TdH - 19Apr2012)

        '''
        Ac = Ac * self.amplitude_to_normalized16ch_conversion
        #variable resistor on carrier chain (mezz rev2 and 3)
        R = [2000.0, 820.0, 200.0, 0.0]
        Eff = [0.9030, 0.8876, 0.8718, 0.8439]

        #transfer functions
        V = Ac * 1.5972e-3 / (100.0+R[Gcarrier])

        #multiplexing factor
        if N==8:
            V*=2.0
        elif N==16:
            pass
        else:
            return -1

        if correct_for_efficiency:
          return V * Eff[Gcarrier]
        else:
          return V


    def theo_VbiasToAc(self,V,Gcarrier,N=16, correct_for_efficiency=False):
        '''
        Theoretical convertion from V_rms across bias resistor to
            fractionnal carrier amplitude from schematics
        Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        V is the V_rms across the bias resistor
        Gcarrier is the carrier gain (0-3)
        N is the multiplexing factor (8 or 16)
        returns the carrier fractionnal amplitude
        '''
        #variable resistor for carrier gain
        R=[2000.0, 820.0, 200.0, 0.0]
        Eff = [0.9030, 0.8876, 0.8718, 0.8439]

        #transfer function
        Ac = V * (100.0+R[Gcarrier]) / 1.5972e-3

        #multiplexing factor
        if N==8:
            Ac/=2.0
        elif N==16:
            pass
        else:
            return -1

        Ac = Ac / self.amplitude_to_normalized16ch_conversion

        if correct_for_efficiency:
          return Ac / Eff[Gcarrier]
        else:
          return Ac


    def DDStoVbias(self, Carrier_amplitude=1, Carrier_gain=0, firmware_version='16ch'):
        '''Converts the carrier gain settings to the rms voltage bias
        across a bolometer.  This function does not take strays
        into account.

        This is a measured transfer function
        Function written by Hannes in ~2008
        '''
        Carrier_amplitude = Carrier_amplitude * self.amplitude_to_normalized16ch_conversion

        if Carrier_gain == 0:
          g = 2.48
        elif Carrier_gain == 1:
          g = 5.56
        elif Carrier_gain == 2:
          g = 16.77
        elif Carrier_gain == 3:
          g = 48.47
        else:
          raise Exception("wtl_ConvertUtils.convert_squid.DDStoVbias given invalid input. Carrier gain must be either 0,1,2,3. Got %s"%(str(Carrier_gain)))

        if firmware_version=='16ch':
          z = 0.5
        elif firmware_version=='8ch':
          z=1
        else:
          raise Exception ("wtl_ConvertUtils.convert_squid.DDStoVbias given invalid input. firmware_version must be '8ch' or '16ch'. Got %s"%(str(firmware_version)))
          return False

        g0 = 2.48
        Rb = .03
        R_current = 210 #includes dewar stray resistance

        Vbias = 9.939e-3*Rb/R_current*g/g0*Carrier_amplitude*z
        return Vbias

class convert_demod:
    def __init__(self, dfmuxApi=None):
      """
          Demod / ADC conversion utilities

          Arguments:
            dfmuxApi - This argument allows the user to choose which API is used and the bit width
                       that is used. The options are as follows:
                        - True
                            The standard "dfmux" API. (24 bit numbers)
                        - False
                            The classic API (wtl_FpgaMotherboard, wtl_DataStreamer, wtl_SquidControl)
                        - 'parser'
                            Data taken from the parser. Like "True" except the numbers are 32 bits wide.
                        - 'dfmux'
                            Equivalent to True
                        - 'classic'
                            Equivalent to False
                        - 'gcp'
                            Equivalent to 'parser'
                        - None
                            Prints a warning message, otherwise identical to 'classic' or False
                        - <integer>
                            Provide the LSB bit shift relative to a 16 bit output.
                            e.g. providing 8 is equivalent to True or 'dfmux'
                            providing 16 is equivalent to 'parser' or 'gcp'
      """
      if dfmuxApi==None:
        print("""
              convert_demod was not called with the dfmuxApi flag.
              Please specify whether you're using the old API (wtl_DataStreamer) with:
              convert_demod(dfmuxApi=False)
              or the new API (libdfmux) with:
              convert_demod(dfmuxApi=True)
              or the parsed (e.g. 32 bit GCP or parser) data with:
              convert_demod(dfmuxApi='parser')

              For backwards compatibility, I'll assume the old API (wtl_DataStreamer)
              and not correct for DAN transfer function.
              Please change your code to explicitly specify the dfmuxApi flag, though.
              """)
        self._to_classic_conversion = 1.
      elif dfmuxApi==True or dfmuxApi=='dfmux':
        self._to_classic_conversion = 1./(2.**(8)) * 0.47685/0.5
      elif dfmuxApi==False or dfmuxApi=='classic':
        self._to_classic_conversion = 1.
      elif dfmuxApi=='parser' or dfmuxApi=='gcp':
        self._to_classic_conversion = 1./(2.**(16)) * 0.47685/0.5
      else:
        self._to_classic_conversion = 1./(2.**(dfmuxApi)) * 0.47685/0.5

    def _rfb_to_float(self, R_FB):
      if R_FB == 'openloop':
        feedback_resistor_ohms = 30600. # this is not actually a feedback resistor,
                                          # but rather a hack to give the correct transfer function
      elif R_FB == '10K':
        feedback_resistor_ohms = 10000.
      elif R_FB == '5K':
        feedback_resistor_ohms = 5000.
      elif R_FB == '3.3K':
        feedback_resistor_ohms = 3333.
      elif (R_FB > 500) and (R_FB < 900000): # numerical value in ohms
        feedback_resistor_ohms = float(R_FB)
      else:
          raise Exception('convert_demod._rfb_to_float :: unable to parse R_FB, which was given as '+str(R_FB))
      return feedback_resistor_ohms        
        
    def convert_rawdump(self, raw_adc_data, Gdemod, R_FB):
        '''
        Convert a rawdump (pre-demodulation timestream from the
        25 MHz, 14 bit ADC) to physical current at the SQUID input
        coil.
        7Nov2011 - TdH
        '''
        Y = [10000.,2000.,200.,0.][Gdemod] + 10.
        current_A = raw_adc_data * (1.
                                    / 2.**14 # 14 bit ADC
                                    * 2. # volts at the ADC
                                    * (50. + 100. + 50.)/100. # voltage divider before ADC
                                    * (50. + Y + 50)/10000. # gain of second stage amplifier
                                    * (50. + 50. + 50. + 50.)/(200. + 200.) # gain of first stage amplifier
                                    * (100.+82.5+1./(1./10. + 1./121.))/(500. + 500.) * (1./self._rfb_to_float(R_FB)) # voltage to current at SQUID coil
                                    * 1.13 # Franky's efficiency factor
                                    )
        return current_A

    def convert_to_volts_dc(self, raw_data, Gdemod=0, dc_bypass_resistors=5000.):
        '''
        Take data from a board that has been modified to read at DC. This assumes you are
        looking at an I channel initialized at zero frequency.

        The defaults (Gdemod=0, dc_bypass_resistors=5000. are for the sptpol calibrator readout
        board.
        '''
        data = raw_data * self._to_classic_conversion

        Y = [10000.,2000.,200.,0.][Gdemod] + 10.
        V_input = data * (1.
                          / 2.**16 # 16 bit numbers in the classic API
                          * 2. # volts at the ADC
                          * (50. + 100. + 50.)/100. # voltage divider before ADC
                          * (50. + Y + 50)/10000. # gain of second stage amplifier
                          * (50. + 50. + 2.*dc_bypass_resistors)/(200. + 200.) # gain of first stage amplifier
                          )

        return V_input

    def theo_AdcToIbolo(self,ADC,Gdemod,R_FB,f_carrier_equals_f_demod,L=10000.):
        '''
        Theoretical convertion ADC to A_rms at bolometers from schematics
        Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        As of August 2011, this function is know to be off.  More measurements
        need to be taken more precisely, but at the moment, a factor of 1.13
        needs to be multiplied to the output of this function to get the
        \'mesured\' transfer function (current is actually higher by 13% then
        what this function reports.

        ADC is the ADC count number (either a pylab array or a scalar
        Gdemod is the demod gain (0-3)
        R_FB is the feedback resistor (10000., 5000. or 3333.)
        f_carrier_equals_f_demod is a boolean : True if f_c==f_d
                                                False if not f_c==f_d
        L is the loopgain, default assumes high loopgain (0.01% effect)
        returns the current in the bolometer in A_rms
        '''
        #variable resistor on demod chain (mezz rev2 and 3)
        ADC *= self._to_classic_conversion
        R=[10000.0, 2000.0, 200.0, 0.0]
          
        G_DF=1.9074 #1.6678=gain of square mixer, 1.9074=gain of semi-sine mixer
        #transfer function
        I = ADC * ((100.0+R[Gdemod]) * 2.3405e-9) / (self._rfb_to_float(R_FB) * G_DF) *abs((1-L)/L)

        #RMS and peak are different only if fc==fd
        if f_carrier_equals_f_demod:
            I=I/2**0.5

        return I


    def theo_IboloToAdc(self,I,Gdemod,R_FB,f_carrier_equals_f_demod,L=10000.):
        '''
        Theoretical convertion A_rms at bolometers to ADC from schematics
        Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        I is the bolometer current in A_rms (either a pylab array or a
            scalar)
        Gdemod is the demod gain (0-3)
        R_FB is the feedback resistor (10000., 5000. or 3333.)
        f_carrier_equals_f_demod is a boolean : True if f_c==f_d
                                                False if not f_c==f_d
        L is the loopgain, default assumes high loopgain (0.01% effect)
        f_carrier_equals_f_demod is a boolean : True if f_c==f_d
                                                    False if not f_c==f_d
        returns the ADC count number
        '''
        #variable resistor on demod chain (mezz rev2 and 3)
        R=[10000.0, 2000.0, 200.0, 0.0]

        G_DF=1.90742 #1.6678=gain of square mixer, 1.9074=gain of semi-sine mixer
        #transfer function
        ADC = I * (self._rfb_to_float(R_FB) * G_DF) / ((100.0+R[Gdemod]) * 2.3405e-9) /abs((1-L)/L)

        #RMS and peak are different only if fc==fd
        if f_carrier_equals_f_demod:
            ADC=ADC*2**0.5

        ADC /= self._to_classic_conversion

        return ADC


    def theo_AdcToVopenloop(self,ADC,fd,Gdemod,f_carrier_equals_f_demod):
        '''
        Converts ADC counts to voltage at the input of the first stage
            amplifier of the SQUID controller board from schematics
        Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        x is the ADC count number
        fd is the demodulated frequency
        Gdemod is the demod gain (0-3)
        f_carrier_equals_f_demod is a boolean : True if f_c==f_d
                                                    False if not f_c==f_d
        returns V_rms at the input of the 1st stage amp of the SQUID ctrl
        '''
        ADC *= self._to_classic_conversion
        #variable resistor in 2nd stage ampifier of demod
        R=[10000.0, 2000.0, 200.0, 0.0]
        Y=R[Gdemod]

                #ADC counts
        V=ADC
        #Digital filter gains
        V=V/1.9074
        #ADC to volt conversion
        V=V*2./16384.
        #voltage divider at input of ADC
        V=V*2.
                #2nd stage demod amplifier gain
        V=V*((100.0+Y)/10000.0)
        #1st stage demod amplifier gain
        V=V/2.
        #2nd stage SQUID controller amplifier gain
        V=V*(100.+82.5+(1./(1./10.+1./121.)))/(500.+500.)
                #current divider
        V=V*(121.+10.)/10.
        #SQUID ctrl 1st stage amplifier gain
        V=V*20./69820.
        #1MHz rolloff of the 1st dtage amplifier of the SQUID controller board
        V=V*(1+(fd/1e6)**2)**0.5

                #if demod and carrier frequencies are the same, rms is not == p
        if f_carrier_equals_f_demod:
            V=V/2**0.5

        return V


    def theo_VopenloopToAdc(self,V,fd,Gdemod,f_carrier_equals_f_demod):
        '''
        Converts voltage at the input of the first stage amplifier of the
            SQUID controller board to ADC counts from schematics
        Francois Aubin <francois.aubin@mail.mcgill.ca> July 2009

        V is the voltage at the input of the first stage amplifier of the
            SQUID controller board
            fd is the demodulated frequency
        Gdemod is the demod gain (0-3)
        f_carrier_equals_f_demod is a boolean : True if f_c==f_d
                                                    False if not f_c==f_d
        returns ADC counts
        '''
            #variable resistor in 2nd stage ampifier of demod
        R=[10000.0, 2000.0, 200.0, 0.0]
        Y=R[Gdemod]

                #ADC counts
        ADC=V
        #1MHz rolloff of the 1st dtage amplifier of the SQUID controller board
        ADC=ADC/(1+(fd/1e6)**2)**0.5
        #SQUID ctrl 1st stage amplifier gain
        ADC=ADC*69820./20.
                #current divider
        ADC=ADC*10./(121.+10.)
        #2nd stage SQUID controller amplifier gain
        ADC=ADC*(500.+500.)/(100.+82.5+(1./(1./10.+1./121.)))
        #1st stage demod amplifier gain
        ADC=ADC*2.
        #2nd stage demod amplifier gain
        ADC=ADC*10000./(100.0+Y)
        #voltage divider at input of ADC
        ADC=ADC/2.
        #volt to ADC conversion
        ADC=ADC*16384./2.
        #Digital filter gains
        ADC=ADC*1.9074

                #if demod and carrier frequencies are the same, rms is not = p
        if f_carrier_equals_f_demod:
            ADC=ADC*2**0.5

        ADC /= self._to_classic_conversion

        return ADC

    #######################################################################################
    # Converts ADC counts_DC to current_rms at bolometers and includes transfer functions #
    #######################################################################################
    def ADCtoItxfr(self,x,Gdemod,R_FB,freq):
        x *= self._to_classic_conversion
        R=[10000.0, 2000.0, 200.0, 0.0]
        m=[-0.2696e-7, -0.2691e-7, -0.5229e-7, -1.8598e-7]
        b=[1.0054, 1.0054, 1.0105, 1.0372]

        G_DF=1.90742 #1.6678=gain of square mixer, 1.9074=gain of semi-sine mixer
        rt2 = 1.41421356
        I = x * ((100.0+R[Gdemod]) * 2.3405e-9) / (self._rfb_to_float(R_FB) * G_DF * rt2)
        if freq >= 200000:
            Itxfr = I / (m[Gdemod] * freq + b[Gdemod])
        else:
            Itxfr = I

        return Itxfr

    #########################################################################################
    # Converts current_rms at bolometers into ADC counts_DC and includes transfer functions #
    #########################################################################################
    def ItoADCtxfr(self,I,Gdemod,R_FB,freq):

        R=[10000.0, 2000.0, 200.0, 0.0]
        m=[-0.2696e-7, -0.2691e-7, -0.5229e-7, -1.8598e-7]
        b=[1.0054, 1.0054, 1.0105, 1.0372]

        G_DF=1.90742 #1.6678=gain of square mixer, 1.9074=gain of semi-sine mixer
        rt2 = 1.41421356
        x = I * (self._rfb_to_float(R_FB) * G_DF * rt2) / ((100.0+R[Gdemod]) * 2.3405e-9)
        if freq >= 200000:
            xtxfr = x * (m[Gdemod] * freq + b[Gdemod])
        else:
            xtxfr = x
        x /= self._to_classic_conversion
        return xtxfr

##
##
#######################################################
##
##  T E S T  C A S E S
##
#######################################################
##
##
class ConvertTestCase (unittest.TestCase):

    def setUp(self):
        #print "Setting up WtlTest cases"
        self.module_mb = convert_mb()
        self.module_mezz = convert_mezz()

    def tearDown(self):
        #print "Cleaning up WtlTest cases"
        self.module_mb = None
        self.module_mezz = None

    def test_01_Convert_Temperature(self):
        '''
            Test the temperature conversion. These examples are found in
            Table 5 of the ADT7301 datasheet.
        '''

        trials = {
            0x3b00 : -40.0,
            0x3c40 : -30.0,
            0x3ce0 : -25.0,
            0x3ec0 : -10.0,
            0x3fff : -0.03125,
            0x0000 : 0,
            0x0001 : 0.03125,
            0x0140 : +10,
            0x0320 : +25,
            0x0640 : +50,
            0x0960 : +75,
            0x0c80 : +100,
            0x0fa0 : +125,
            0x12c0 : +150
        }

        for value, expected in trials.items():
            calculated = self.module_mb.temperature(value)
            self.assertEqual(calculated, expected,
                'error converting 0x%08x to mb temperature %f (calculated %f)' %
                (value, expected, calculated))


    def test_02_Convert_MB_neg(self):
        ''' 2: Test the MB Negative conversion
        '''
        self.assertEqual(self.module_mb.neg_voltage(1520), -7.283203125, 'error converting mb neg voltage')
        self.assertEqual(self.module_mb.neg_voltage(2520), 9.6357421875, 'error converting mb neg voltage')

    def test_03_Convert_MB_pos(self):
        ''' 2: Test the MB Positive conversion
        '''
        self.assertEqual(self.module_mb.pos_voltage(1024), 3.2999999999999998, 'error converting mb pos voltage')
        self.assertEqual(self.module_mb.pos_voltage(4096), 13.199999999999999, 'error converting mb pos voltage')

    def test_04_Convert_Mezz_neg(self):
        ''' 2: Test the Mezz Negative conversion
        '''
        self.assertEqual(self.module_mezz.neg_voltage(1768, 2020), -8.5561523437499929, 'error converting mezz neg voltage')
        self.assertEqual(self.module_mezz.neg_voltage(1224, 1520), -3.8671874999999929, 'error converting mezz neg voltage')

    def test_05_Convert_Mezz_pos(self):
        ''' 2: Test the Mezz Positive conversion
        '''
        self.assertEqual(self.module_mezz.pos_voltage(1024), 2.4749999999999996, 'error converting mezz pos voltage')
        self.assertEqual(self.module_mezz.pos_voltage(4096), 9.8999999999999986, 'error converting mezz pos voltage')


def suite():
    suite = unittest.makeSuite(ConvertTestCase,'test')
    return suite

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from spt3g import core
import time, sys, glob, os

# Clean up from previous run
for f in glob.glob('multitest*.g3'):
	os.remove(f)

# File to disk
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=1000)
n = 0
def addinfo(fr):
	global n
	if fr.type != core.G3FrameType.Timepoint:
		return
	fr['time'] = core.G3Time(int(time.time()*core.G3Units.s))
	fr['count'] = n
	n += 1
pipe.Add(addinfo)

class checkwriter(core.G3MultiFileWriter):
	def __call__(self, frame):
		if frame.type == core.G3FrameType.EndProcessing:
			assert os.path.exists(self.current_file)
		super().__call__(frame)

pipe.Add(lambda fr: fr.type != core.G3FrameType.PipelineInfo) # Avoid extra frames that complicate accounting
pipe.Add(checkwriter, filename='multitest-%02u.g3', size_limit=20*1024)
pipe.Add(checkwriter, filename=lambda frame,seq: 'multitest2-%02d.g3' % seq, size_limit=20*1024)
pipe.Add(checkwriter, filename='multitest3-%02u.g3', size_limit=20*1024, divide_on=[core.G3FrameType.Timepoint])
pipe.Add(checkwriter, filename='multitest4-%02u.g3', size_limit=2000000*1024, divide_on=lambda fr: fr['count'] % 200 == 0)

pipe.Run()

# Check that various ways of splitting produce the right number of files
nfiles = len(glob.glob('multitest-*.g3'))
print('Seeing %d files, expecting 7 from format string constructor' % nfiles)
assert(nfiles == 7)

nfiles = len(glob.glob('multitest2-*.g3'))
print('Seeing %d files, expecting 7 from callback constructor' % nfiles)
assert(nfiles == 7)

nfiles = len(glob.glob('multitest3-*.g3'))
print('Seeing %d files, expecting 1000 from boundary division constructor' % nfiles)
assert(nfiles == 1000)

nfiles = len(glob.glob('multitest4-*.g3'))
print('Seeing %d files, expecting 5 from boundary callback constructor' % nfiles)
assert(nfiles == 5)

# Check we can read them all
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename=sorted(glob.glob('multitest4-*.g3')))
m = 0
def countframes(fr):
	global m
	if fr.type != core.G3FrameType.Timepoint:
		return
	assert(fr['count'] == m)
	m += 1
pipe.Add(countframes)
pipe.Run()
print('Read %d/%d frames' % (m,n))
assert(m == n)

# Clean up
for f in glob.glob('multitest*.g3'):
	os.remove(f)




from .. import core
from . import DfMuxWiringMap, DfMuxChannelMapping
import struct, socket

'''
Code to generate wiring map objects (and wiring frames in general) from
control system hardware maps.
'''

@core.indexmod
class GenerateFakeHardwareMap(object):
    '''Inserts a fake hardware map into the data stream. Takes a wiring frame as an argument to the constructor, which it will inject before any other frames.'''
    def __init__(self, frame):
        self.ran = False
        self.insert = frame
    def __call__(self, frame):
        if self.ran:
            return frame
        else:
            self.ran = True
            return [self.insert, frame]

NUM_MEZZANINES = 2
NUM_MODS_PER_MEZZ = 4
NUM_MODULES = NUM_MEZZANINES * NUM_MODS_PER_MEZZ

@core.indexmod
class PyDfMuxWiringMapInjector(object):
    '''
    Insert a wiring map derived from a pydfmux hardware map into the data
    stream ahead of what would otherwise be the first frame.
   
    Optionally filter for detectors described by the mask in <pathstring>
    (see pydfmux documentation for hwm.channel_maps_from_pstring()) and
    detectors in one of the states identified by the state argument.
    '''
    def __init__(self, pydfmux_hwm, pathstring=None, state=[]):
        self.ran = False

        from pydfmux.core import dfmux as pydfmux
        self.hwmf = core.G3Frame(core.G3FrameType.Wiring)
        hwm = DfMuxWiringMap()
        
        if pathstring:
            chan_map_query = pydfmux_hwm.channel_maps_from_pstring(pathstring)
        else:
            chan_map_query = pydfmux_hwm.query(pydfmux.ChannelMapping)

        if len(state) > 0:
            for bolo in pydfmux_hwm.query(pydfmux.Bolometer):
                if bolo.readout_channel:
                    bolo.state = bolo.retrieve_bolo_state().state
            pydfmux_hwm.commit()
            chan_map_query = chan_map_query.join(pydfmux.ChannelMapping, pydfmux.Bolometer).filter(pydfmux.Bolometer.state._in(state))

        for bolo in chan_map_query:
            mapping = DfMuxChannelMapping()
            mapping.board_ip = struct.unpack("i", socket.inet_aton(socket.gethostbyname('iceboard' + str(bolo.iceboard.serial) + '.local')))[0]
            mapping.board_serial = int(bolo.iceboard.serial)
            mapping.board_slot = bolo.iceboard.slot if bolo.iceboard.slot else -1
            mapping.crate_serial = int(bolo.iceboard.crate.serial) if bolo.iceboard.slot else -1
            mezz = bolo.readout_channel.mezzanine.mezzanine
            mod = bolo.readout_channel.module.module
            mapping.module = (mezz - 1) * NUM_MODS_PER_MEZZ + (mod - 1)
            mapping.channel = bolo.readout_channel.channel - 1 # pydfmux HWMs use 1-indexing of channels, while FPGA uses 0-indexing
            hwm[str(bolo.bolometer.name)] = mapping
        self.hwmf['WiringMap'] = hwm
        self.hwmf['ReadoutSystem'] = 'ICE'

        try:
            from pydfmux import current_transferfunction
            self.hwmf['DfMuxTransferFunction'] = current_transferfunction
        except ImportError:
            pass

    def __call__(self, frame):
        if self.ran:
            return frame

        self.ran = True
        out = [self.hwmf, frame]
        del self.hwmf
        return out

# Compatibility
PyDfMuxHardwareMapInjector = PyDfMuxWiringMapInjector


@core.usefulfunc
def PathStringForBolo(wiringmap, bolo):
    '''
    Obtain the channel pathstring for a bolometer named "bolo"
    using the passed wiring map.
    '''
    wiringentry = wiringmap[bolo]
    if wiringentry.crate_serial >= 0:
        path = '/'.join([
            '%03d' % wiringentry.crate_serial,
            '%d' % wiringentry.board_slot,
        ])
    else:
        path = '%04d' % wiringentry.board_serial
    path = '/'.join([
        path,
        '%d' % ((wiringentry.module // NUM_MODULES) + 1),
        '%d' % ((wiringentry.module % NUM_MODULES) + 1),
        '%d' % (wiringentry.channel + 1),
    ])
    return path


@core.indexmod
def PyDfMuxBolometerPropertiesInjector(frame, pydfmux_hwm=None, angle_per_mm = 4.186*core.G3Units.deg/1000):
    '''
    Insert a calibration frame following any wiring frame containing a
    BolometerPropertiesMap named "NominalBolometerProperties" that has
    the properties of each bolometer as defined by the given pydfmux 
    hardware map.
    '''

    if frame.type != core.G3FrameType.Wiring:
        return

    from .. import calibration
    from pydfmux.core import dfmux as pydfmux

    cal = core.G3Frame(core.G3FrameType.Calibration)
    bpm = calibration.BolometerPropertiesMap()

    for bolo in pydfmux_hwm.query(pydfmux.Bolometer):
        bp = calibration.BolometerProperties()
        if hasattr(bolo, 'physical_name'):
            bp.physical_name = str(bolo.wafer.name) + '_' + str(bolo.physical_name)
        if hasattr(bolo, 'x_mm') and bolo.x_mm is not None:
            bp.x_offset = bolo.x_mm * angle_per_mm
        if hasattr(bolo, 'y_mm') and bolo.y_mm is not None:
            bp.y_offset = bolo.y_mm * angle_per_mm
        if bolo.wafer is not None:
            bp.wafer_id = str(bolo.wafer.name)
        if hasattr(bolo, 'pixel') and bolo.pixel is not None:
            bp.pixel_id = str(bolo.pixel)
        if hasattr(bolo, 'pixel_type') and bolo.pixel_type is not None:
            bp.pixel_type = str(bolo.pixel_type)
        if hasattr(bolo, 'observing_band') and bolo.observing_band is not None:
            bp.band = float(bolo.observing_band)*core.G3Units.GHz
        if hasattr(bolo, 'polarization_angle') and bolo.polarization_angle is not None:
            bp.pol_angle = float(bolo.polarization_angle)*core.G3Units.deg
            bp.pol_efficiency = 1.0
        if hasattr(bolo, 'coupling') and bolo.coupling is not None:
            if bolo.coupling == "optical":
                bp.coupling = calibration.BolometerCouplingType.Optical
            elif bolo.coupling == "dark_tres":
                bp.coupling = calibration.BolometerCouplingType.DarkTermination
            elif bolo.coupling == "dark_xover":
                bp.coupling = calibration.BolometerCouplingType.DarkCrossover
            elif bolo.coupling == "resistor":
                bp.coupling = calibration.BolometerCouplingType.Resistor
            else:
                raise ValueError("Unrecognized coupling type %s" % bolo.coupling)

        bpm[str(bolo.name)] = bp

    cal['NominalBolometerProperties'] = bpm
    from pydfmux import current_transferfunction
    cal['DfMuxTransferFunction'] = current_transferfunction

    return [frame, cal]

core.indexmod
class PyDfMuxWiringMapInjectorAllChannels(object):
    '''
    Insert a wiring map derived from a pydfmux hardware map, but
    ignoring the channel mappings and adding all the channels on each defined
    module from the HWM.
    '''
    def __init__(self, pydfmux_hwm, channels=64):
        '''
        Use readout modules from provided pydfmux_hwm and add channels to it
        up to the index specified in channels.
        '''
        self.ran = False
        self.hwm = pydfmux_hwm
        self.channels = channels
    def __call__(self, frame):
        if self.ran:
            return frame

        from pydfmux.core import dfmux as pydfmux
        hwmf = core.G3Frame(core.G3FrameType.Wiring)
        hwm = DfMuxWiringMap()
        for mod in self.hwm.query(pydfmux.ReadoutModule):
            for bolo in range(self.channels):
                mapping = DfMuxChannelMapping()
                mapping.board_ip = struct.unpack("i", socket.inet_aton(socket.gethostbyname('iceboard' + str(mod.iceboard.serial) + '.local')))[0]
                mapping.board_serial = int(mod.iceboard.serial)
                mapping.board_slot = mod.iceboard.slot if mod.iceboard.slot else -1
                mapping.crate_serial = int(mod.iceboard.crate.serial) if mod.iceboard.slot else -1
                mezz = mod.mezzanine.mezzanine
                # pydfmux HWMs use 1-indexing of modules, while FPGA uses 0-indexing
                mapping.module = (mezz - 1) * NUM_MODS_PER_MEZZ + mod.module - 1
                mapping.channel = bolo
                hwm[str(mapping)] = mapping
        hwmf['WiringMap'] = hwm
        hwmf['ReadoutSystem'] = 'ICE'
        self.ran = True

        return [hwmf, frame]

# Compatibility
PyDfMuxHardwareMapInjectorAllChannels = PyDfMuxWiringMapInjectorAllChannels

@core.indexmod
class DfmlHardwareMapInjector(object):
    def __init__(self, dfml_hwm):
        self.ran = False
        self.dfml_hwm = dfml_hwm

        self.hwm = DfMuxWiringMap()
        for ch in self.dfml_hwm.channels:
            bolo_id = str(self.dfml_hwm(ch, 'channel_id'))
            if self.dfml_hwm(ch, 'bolo_id') == '':
                # Check bolo_id, which is set iff this is a channel we care about
                continue
            mapping = DfMuxChannelMapping()
            mapping.board_ip = struct.unpack("i", socket.inet_aton( self.dfml_hwm(ch, 'dfmux_ip') ))[0]
            mapping.board_serial = int(self.dfml_hwm(ch, 'dfmux_id'))
            if self.dfml_hwm(ch, 'crate_slot') is not None:
                mapping.board_slot = int(self.dfml_hwm(ch, 'crate_slot'))
            mapping.crate_serial = -1
            mapping.module = self.dfml_hwm(ch, 'module') - 1
            mapping.channel = self.dfml_hwm(ch, 'chan_num') - 1
            self.hwm[bolo_id] = mapping
    def __call__(self, frame):
        if self.ran:
            return frame
        hwmf = core.G3Frame(core.G3FrameType.Wiring)
        hwmf['WiringMap'] = self.hwm
        hwmf['ReadoutSystem'] = 'DfMux'
        self.ran = True
        return [hwmf, frame]




import numpy as np
import os

from .. import core

__all__ = [
    "convert_azel_to_radec",
    "convert_radec_to_azel",
    "convert_radec_to_gal",
    "convert_gal_to_radec",
    "LocalToAstronomicalPointing",
    "EquatorialToGalacticPointing",
]


def get_location(location="spt"):
    """
    Return the astropy EarthLocation for the given location name.

    Arguments
    ---------
    location : str or EarthLocation instance
        If a string, must be a recognized location name.  Currently
        only "spt" is supported.
    """
    from astropy.coordinates import EarthLocation
    import astropy.units

    if isinstance(location, EarthLocation):
        return location

    location = str(location).lower()
    if location == "spt":
        return EarthLocation(
            lat=-89.991066 * astropy.units.deg,
            lon=-44.65 * astropy.units.deg,
            height=2835.0 * astropy.units.meter,
        )

    raise NotImplementedError


iers_checked = False


def check_iers(mjd):
    """
    Check whether IERS calculations will work, and load an IERS database file
    from backup if all else fails.

    Arguments
    ---------
    mjd : Array-like
        MJD timestamps for which an IERS calculation must be computed.

    Returns
    -------
    t : astropy.time.Time instance
        Time instance for the input MJD(s).
    """
    from astropy.utils import iers
    import astropy.time

    global iers_checked
    if iers_checked:
        return astropy.time.Time(mjd, format="mjd")

    try:
        if os.getenv("SPT3G_IERS_AUTO_URL"):
            iers.conf.iers_auto_url = os.getenv("SPT3G_IERS_AUTO_URL")
        if os.getenv("SPT3G_IERS_REMOTE_TIMEOUT"):
            iers.conf.remote_timeout = float(os.getenv("SPT3G_IERS_REMOTE_TIMEOUT"))
    except:
        pass

    mjd1 = np.atleast_1d(mjd)[-1]
    t1 = astropy.time.Time(mjd1, format="mjd")

    # check if accessing the IERS table outright works.
    try:
        t1.ut1
        iers_checked = True
        return astropy.time.Time(mjd, format="mjd")
    except:
        pass

    # if that fails, allow extrapolation
    iers.conf.auto_max_age = None
    t1 = astropy.time.Time(mjd1, format="mjd")
    try:
        t1.ut1
        core.log_warn("IERS auto update failed, allowing extrapolation", unit="IERS")
        iers_checked = True
        return astropy.time.Time(mjd, format="mjd")
    except:
        pass

    # and if that fails, use a locally cached file that is hopefully setup correctly.
    fname = os.path.join(os.path.dirname(os.path.abspath(__file__)), "finals2000A.all")
    iers.conf.auto_download = False
    iers.IERS.iers_table = iers.IERS_A.open(fname)
    t1 = astropy.time.Time(mjd1, format="mjd")
    t1.ut1
    core.log_warn("Using IERS table from local cache {}".format(fname), unit="IERS")
    iers_checked = True
    return astropy.time.Time(mjd, format="mjd")


def convert_deg(d, system="g3"):
    """
    Convert the input array to the appropriate system of units.

    Arguments
    ---------
    d : array_like
        Data vector
    system : str
        System to convert to, either "g3" or "astropy".  Assumes conversion from
        the other system.
    """
    import astropy.units

    system = str(system).lower()
    if system == "astropy":
        return np.asarray(d) / core.G3Units.deg * astropy.units.deg
    elif system == "g3":
        return np.asarray(d / astropy.units.deg) * core.G3Units.deg
    raise NotImplementedError


@core.usefulfunc
def convert_azel_to_radec(az, el, location="spt", mjd=None):
    """
    Convert timestreams of local azimuth and elevation to right ascension and
    declination.

    Arguments
    ---------
    az, el : np.ndarray or G3Timestream
        Array of local coordinates. If inputs are G3Timestream objects,
        G3Timestreams are also returned.
    location : str or astropy.coordinates.EarthLocation instance
        The telescope location on Earth.
    mjd : np.ndarray
        An array of times for each az/el sample.  If input az and el
        are not G3Timestreams, this argument is required.

    Returns
    -------
    ra, dec : np.ndarray or G3Timestream
    """
    import astropy.coordinates

    singleton = False
    if isinstance(az, core.G3Timestream):
        assert az.start == el.start
        assert az.stop == el.stop
        assert az.n_samples == el.n_samples
        mjd = np.asarray([i.mjd for i in az.times])
    else:
        try:
            len(az)
        except TypeError:
            singleton = True
        if singleton:
            az = np.atleast_1d(az)
            el = np.atleast_1d(el)
            mjd = np.atleast_1d(mjd)
        assert len(az) == len(el)

    t = check_iers(mjd)

    # record locations of bad elevation values to mark them later
    badel_inds = np.where(
        (el < -90.0 * core.G3Units.deg) | (el > 90.0 * core.G3Units.deg)
    )
    el[badel_inds] = 0.0 * core.G3Units.deg

    k = astropy.coordinates.AltAz(
        az=convert_deg(az, "astropy"),
        alt=convert_deg(el, "astropy"),
        obstime=t,
        location=get_location(location),
        pressure=0,
    )
    kt = k.transform_to(astropy.coordinates.FK5())

    ra = convert_deg(kt.ra, "g3")
    dec = convert_deg(kt.dec, "g3")
    dec[badel_inds] = np.nan

    if isinstance(az, core.G3Timestream):
        ra = core.G3Timestream(ra)
        dec = core.G3Timestream(dec)
        ra.start = dec.start = az.start
        ra.stop = dec.stop = az.stop
    elif singleton:
        ra = ra[0]
        dec = dec[0]

    return (ra, dec)


@core.usefulfunc
def convert_radec_to_azel(ra, dec, location="spt", mjd=None):
    """
    Convert timestreams of right ascension and declination to local
    azimuth and elevation.

    Arguments
    ---------
    ra, dec : np.ndarray or G3Timestream
        Array of Equatorial sky coordinates. If inputs are G3Timestream
        objects, G3Timestreams are also returned.
    location : str or astropy.coordinates.EarthLocation instance
        The telescope location on Earth.
    mjd : np.ndarray
        An array of times for each ra/dec sample.  If input ra and dec
        are not G3Timestreams, this argument is required.

    Returns
    -------
    az, el : np.ndarray or G3Timestream
    """
    import astropy.coordinates

    singleton = False
    if isinstance(ra, core.G3Timestream):
        assert ra.start == dec.start
        assert ra.stop == dec.stop
        assert ra.n_samples == dec.n_samples
        mjd = np.asarray([i.mjd for i in ra.times])
    else:
        try:
            len(ra)
        except TypeError:
            singleton = True
        if singleton:
            ra = np.atleast_1d(ra)
            dec = np.atleast_1d(dec)
            mjd = np.atleast_1d(mjd)
        assert len(ra) == len(dec)

    t = check_iers(mjd)

    k = astropy.coordinates.FK5(
        ra=convert_deg(ra, "astropy"), dec=convert_deg(dec, "astropy"),
    )
    kt = k.transform_to(
        astropy.coordinates.AltAz(
            obstime=t,
            location=get_location(location),
            pressure=0,
        )
    )

    az = convert_deg(kt.az, "g3")
    el = convert_deg(kt.alt, "g3")

    if isinstance(ra, core.G3Timestream):
        az = core.G3Timestream(az)
        el = core.G3Timestream(el)
        az.start = el.start = ra.start
        az.stop = el.stop = ra.stop
    elif singleton:
        az = az[0]
        el = el[0]

    return (az, el)


@core.usefulfunc
def convert_radec_to_gal(ra, dec):
    """
    Convert timestreams of right ascension and declination to Galactic
    longitude and latitude.

    Arguments
    ---------
    ra, dec : np.ndarray or G3Timestream
        Array of Equatorial sky coordinates. If inputs are G3Timestream
        objects, G3Timestreams are also returned.

    Returns
    -------
    glon, glat : np.ndarray or G3Timestream
    """
    import astropy.coordinates

    singleton = False
    if isinstance(ra, core.G3Timestream):
        assert ra.start == dec.start
        assert ra.stop == dec.stop
        assert ra.n_samples == dec.n_samples
    else:
        try:
            len(ra)
        except TypeError:
            singleton = True
        if singleton:
            ra = np.atleast_1d(ra)
            dec = np.atleast_1d(dec)
        assert len(ra) == len(dec)

    k = astropy.coordinates.FK5(
        ra=convert_deg(ra, "astropy"), dec=convert_deg(dec, "astropy"),
    )
    kt = k.transform_to(astropy.coordinates.Galactic())

    glon = convert_deg(kt.l, "g3")
    glat = convert_deg(kt.b, "g3")

    if isinstance(ra, core.G3Timestream):
        glon = core.G3Timestream(glon)
        glat = core.G3Timestream(glat)
        glon.start = glat.start = ra.start
        glon.stop = glat.stop = ra.stop
    elif singleton:
        glon = glon[0]
        glat = glat[0]

    return (glon, glat)


@core.usefulfunc
def convert_gal_to_radec(glon, glat):
    """
    Convert timestreams of Galactic longitude and latitude to right ascension
    and declination.

    Arguments
    ---------
    glon, glat : np.ndarray or G3Timestream
        Array of Galactic sky coordinates. If inputs are G3Timestream
        objects, G3Timestreams are also returned.

    Returns
    -------
    ra, dec : np.ndarray or G3Timestream
    """
    import astropy.coordinates

    singleton = False
    if isinstance(glon, core.G3Timestream):
        assert glon.start == glat.start
        assert glon.stop == glat.stop
        assert glon.n_samples == glat.n_samples
    else:
        try:
            len(glon)
        except TypeError:
            singleton = True
        if singleton:
            glon = np.atleast_1d(glon)
            glat = np.atleast_1d(glat)
        assert len(glon) == len(glat)

    k = astropy.coordinates.Galactic(
        l=convert_deg(glon, "astropy"), b=convert_deg(glat, "astropy"),
    )
    kt = k.transform_to(astropy.coordinates.FK5())

    ra = convert_deg(kt.ra, "g3")
    dec = convert_deg(kt.dec, "g3")

    if isinstance(ra, core.G3Timestream):
        ra = core.G3Timestream(ra)
        dec = core.G3Timestream(dec)
        ra.start = dec.start = glon.start
        ra.stop = dec.stop = glon.stop
    elif singleton:
        ra = ra[0]
        dec = dec[0]

    return (ra, dec)


@core.indexmod
def LocalToAstronomicalPointing(
    frame,
    az_timestream="BoresightAz",
    el_timestream="BoresightEl",
    ra_timestream="BoresightRa",
    dec_timestream="BoresightDec",
    Telescope="spt",
):
    """
    Converts a set of timestreams in Scan frames representing Az and El pointing
    of the telescope into RA and Declination timestreams, stored in the frame
    under their respective names.
    """

    if frame.type != core.G3FrameType.Scan:
        return

    ra, dec = convert_azel_to_radec(
        frame[az_timestream], frame[el_timestream], location=Telescope
    )

    frame[ra_timestream] = ra
    frame[dec_timestream] = dec


@core.indexmod
def EquatorialToGalacticPointing(
    frame,
    ra_timestream="BoresightRa",
    dec_timestream="BoresightDec",
    glon_timestream="BoresightGalLon",
    glat_timestream="BoresightGalLat",
):
    """
    Converts a set of timestreams in Scan frames representing RA and Declination
    pointing of the telescope into Galactic longitude and latitude timestreams,
    stored in the frame under their respective names.
    """

    if frame.type != core.G3FrameType.Scan:
        return

    glon, glat = convert_radec_to_gal(frame[ra_timestream], frame[dec_timestream])

    frame[glon_timestream] = glon
    frame[glat_timestream] = glat



#!/usr/bin/env python

from spt3g import core
import time

# File to disk
pipe = core.G3Pipeline()
pipe.Add(core.G3InfiniteSource, type=core.G3FrameType.Timepoint, n=10)
n = 0


def addinfo(fr):
    global n
    if fr.type != core.G3FrameType.Timepoint:
        return
    fr["time"] = core.G3Time(int(time.time() * core.G3Units.s))
    fr["count"] = n
    n += 1


pipe.Add(addinfo)
pipe.Add(core.Dump)
pipe.Add(core.G3Writer, filename="test.g3.bz2", buffersize=1024)
pipe.Run()

# And back from disk
print("Reading")
pipe = core.G3Pipeline()
pipe.Add(core.G3Reader, filename="test.g3.bz2", buffersize=1024)
pipe.Add(core.Dump)
n = 0


def checkinfo(fr):
    global n
    if fr.type != core.G3FrameType.Timepoint:
        return
    if "time" not in fr:
        raise KeyError("time")
    if fr["count"] != n:
        raise ValueError("Out of order frame")
    n += 1


pipe.Add(checkinfo)
pipe.Run()

if n != 10:
    raise ValueError("Wrong number of frames (%d should be %d)" % (n, 10))



import numpy
from .. import core, dfmux

@core.indexmod
class UnpackSPTpolHKData(object):
    '''
    Extracts SPTpol housekeeping information from ARC files
    '''

    def __init__(self):
        self.hk_seen = False

    def __call__(self, f):
        if f.type != core.G3FrameType.GcpSlow:
            return

        hkregs = f['array']['muxHousekeeping']

        # HK registers are just duplicated when unchanged. Deduplicate them here
        if self.hk_seen and not numpy.asarray(hkregs['hkIsValid']).any():
            return
        self.hk_seen = True

        hkout = dfmux.DfMuxHousekeepingMap()

        for i, serial in enumerate(hkregs['moboId']):
            board = dfmux.HkBoardInfo()
            board.timestamp = hkregs['utc'] # TimeStamp is off by exactly 30 days, so ignore it
            board.timestamp_port = 'UNKNOWN'
            board.serial = str(serial)
            board.fir_stage = hkregs['firStage'][i]
            board.voltages['Ext5V5N'] = hkregs['mbExt5V5N'][i]
            board.voltages['Ext5V5P'] = hkregs['mbExt5V5P'][i]
            board.voltages['Int5V5N'] = hkregs['mbInt5V5N'][i]
            board.voltages['Int5V5P'] = hkregs['mbInt5V5P'][i]
            board.voltages['Squid8V5N'] = hkregs['mbSquid8V5N'][i]
            board.voltages['Squid8V5P'] = hkregs['mbSquid8V5P'][i]
            board.voltages['VAuxN'] = hkregs['mbVAuxN'][i]
            board.voltages['VAuxP'] = hkregs['mbVAuxP'][i]
            # XXX: mbSensorX?

            for j, mezz in enumerate(['A', 'B']):
                mezzhk = dfmux.HkMezzanineInfo()
                mezzhk.power = True
                mezzhk.present = True
                mezzhk.serial = 'N/A'
                mezzhk.part_number = 'SPTpol Mezz'
                mezzhk.revision = ''
    
                for v in ['V3P', 'V5N', '_VDemodA', '_VDemodB']:
                    mezzhk.voltages[v.strip('_')] = hkregs['mezz' + mezz + v][i]

                for k, mod in enumerate(['A', 'B']):
                    modhk = dfmux.HkModuleInfo()
                    modhk.module_number = k
                    modhk.carrier_gain = hkregs['carrierGain'][i][j*2 + k]
                    modhk.nuller_gain = hkregs['nullerGain'][i][j*2 + k]
                    modhk.demod_gain = hkregs['demodGain'][i][j*2 + k]
                    if 'mezz' + mezz + 'Car' + mod + 'Overload' in hkregs: 
                        modhk.carrier_railed = hkregs['mezz' + mezz + 'Car' + mod + 'Overload'][i]
                        modhk.nuller_railed = hkregs['mezz' + mezz + 'Nul' + mod + 'Overload'][i]
                    modhk.demod_railed = hkregs['mezz' + mezz + 'Adc' + mod + 'Overload'][i]

                    nchan = len(hkregs['carrierSettings'][0])//(3*2*2) # 3 fields per mezz/module
                    for chan in range(nchan):
                        chanhk = dfmux.HkChannelInfo()
                        chanhk.channel_number = chan+1
                        arcchan = (j*2 + k)*nchan + chan

                        if 'danGain' in hkregs:
                            chanhk.dan_gain = hkregs['danGain'][i][arcchan]
                            chanhk.dan_feedback_enable = hkregs['danFeedbackEnable'][i][arcchan]
                            chanhk.dan_streaming_enable = hkregs['danStreamingEnable'][i][arcchan]
                            chanhk.dan_accumulator_enable = hkregs['danAccumulatorRailed'][i][arcchan]
                            chanhk.dan_railed = hkregs['danAccumulatorRailed'][i][arcchan]

                        # Decrypt xSettings registers: (freq, phase, amp)
                        chanhk.carrier_frequency = hkregs['carrierSettings'][i][arcchan*3 + 0]*25.e6/(2.**32-1)*core.G3Units.Hz
                        chanhk.carrier_amplitude = hkregs['carrierSettings'][i][arcchan*3 + 2]/(2.**23 - 1.)
                        chanhk.nuller_amplitude = hkregs['nullerSettings'][i][arcchan*3 + 2]/(2.**23 - 1.)
                        chanhk.demod_frequency = hkregs['demodSettings'][i][arcchan*2 + 0]*25.e6/(2.**32-1)*core.G3Units.Hz

                        modhk.channels[chan+1] = chanhk

                    mezzhk.modules[k+1] = modhk

                board.mezz[j+1] = mezzhk

            hkout[serial] = board

        hkframe = core.G3Frame(core.G3FrameType.Housekeeping)
        hkframe['DfMuxHousekeeping'] = hkout

        return [f, hkframe]




#!/usr/bin/env python

import numpy as np
from spt3g import core, maps

maplist = [
    maps.FlatSkyMap(np.random.randn(64, 12 * 64), core.G3Units.arcmin),
    maps.HealpixSkyMap(np.random.randn(12 * 64 * 64)),
]

for x in maplist:
    print(x)
    m = x.to_mask()

    # attributes
    assert(m.size == x.size)
    assert(m.shape == x.shape)

    # element-wise assignment
    if isinstance(x, maps.FlatSkyMap):
        v = m[5, 5]
        m[5, 5] = not v
        assert(m[5, 5] != v)
        m[5, 5] = v
        assert(m[5, 5] == v)
    else:
        v = m[5]
        m[5] = not v
        assert(m[5] != v)
        m[5] = v
        assert(m[5] == v)

    # masked assignment and retrieval
    print("assignment")
    pospixels = x[x > 0]
    assert(len(pospixels) == (x > 0).sum())
    assert((np.asarray(pospixels) > 0).all())

    x2 = x.__class__(x)
    x2[x2 > 0] = 0
    assert((x2 > 0).sum() == 0)
    nzero = (x2 == 0).sum()
    x2[x2 == 0] = np.abs(np.random.randn((x2 == 0).sum())) + 1
    assert((x2 > 0).sum() == nzero)
    assert((x2 == 0).sum() == 0)

    # init and clone
    mx = maps.G3SkyMapMask(x, True)
    assert((mx == m).all())
    mn = maps.G3SkyMapMask(x, False)
    assert(mn.sum() == 0)
    a = np.asarray(x).ravel()
    ma = m.array_clone(a)
    assert((m == ma).all())
    mp = m.array_clone(a > 0)
    assert(((x > 0) == mp).all())

    # float comparison
    print("float comp")
    m1 = x == 0
    m2 = x != 0
    assert(m1.sum() + m2.sum() == x.size)

    m1 = x <= 0
    m2 = x > 0
    assert(m1.sum() + m2.sum() == x.size)

    m1 = x < 0
    m2 = x >= 0
    assert(m1.sum() + m2.sum() == x.size)

    # map comparison
    print("map comp")
    y = 2 * x.clone()

    m1 = y == x
    m2 = y != x
    assert(m1.sum() + m2.sum() == x.size)

    m1 = y <= x
    m2 = y > x
    assert(m1.sum() + m2.sum() == x.size)

    m1 = y < x
    m2 = y >= x
    assert(m1.sum() + m2.sum() == x.size)

    # logic operators
    print("logic")
    assert((m1 == ~m2).all())
    assert(not (m1 & m2).any())
    assert((m1 | m2).all())
    assert((m1 ^ m2).all())

    m3 = m1.clone()
    m3 |= m2
    assert(m3.all())
    m3 = m1.clone()
    m3 &= m2
    assert(not m3.any())
    m3 = m1.clone()
    m3 ^= m2
    assert(m3.all())
    m3 = m1.clone()
    m3.invert()
    assert((m3 == m2).all())

    # convert to mask
    print("conversion")
    x2 = x.clone()
    x2 *= m1
    assert((x2.to_mask() == m1).all())
    assert(((x * m1).to_mask() == m1).all())

    x3 = x.clone()
    x3.apply_mask(m1, inverse=True)
    assert((x3.to_mask() == m2).all())

    x4 = m1.to_map()
    assert(np.sum(x4) == m1.sum())
    assert((x4.to_mask() == m1).all())

    mpix = m1.nonzero()
    xpix, _ = x4.nonzero_pixels()
    xpixm = x4.nonzero()

    assert(len(set(mpix) ^ set(xpix)) == 0)
    assert(len(set(mpix) ^ set(xpixm)) == 0)

    m1.apply_mask(m2)
    assert(not m1.any())

    # Direct array conversion
    assert(np.asarray(m2).shape == np.asarray(m2.to_map()).shape)
    assert((np.asarray(m2) == np.asarray(m2.to_map())).all())

    # ufuncs
    print("ufuncs")
    if isinstance(x, maps.FlatSkyMap):
        x.sparse = True
    else:
        x.ringsparse = True
    m = x < 0

    for attr in ["all", "any", "sum", "mean", "var", "std", "min", "max", "argmin", "argmax"]:
        for w in [None, m]:
            if w is not None and attr in ["argmin", "argmax"]:
                continue
            kwargs = {"where": w} if w is not None else {}
            if attr in ["std", "var"]:
                kwargs["ddof"] = 1
            try:
                v1 = getattr(np, attr)(x, **kwargs)
            except TypeError as e:
                # ignore errors with older numpy versions
                if w is not None and "unexpected keyword argument 'where'" in str(e):
                    continue
                raise
            v2 = getattr(x, attr)(**kwargs)
            if w is not None:
                kwargs["where"] = np.asarray(w.to_map(), dtype=bool)
                if attr in ["min", "max"]:
                    kwargs["initial"] = np.inf if attr == "min" else -np.inf
            v3 = getattr(np, attr)(np.asarray(x.copy()), **kwargs)
            print(attr, w, v1, v2, v3)
            assert(v1 == v2)
            assert(np.isclose(v2, v3))
            if isinstance(x, maps.FlatSkyMap):
                assert(x.sparse == True)
            else:
                assert(x.ringsparse == True)

    x[m.nonzero()[0]] = np.nan

    for attr in ["nansum", "nanmean", "nanvar", "nanstd", "nanmin", "nanmax", "nanargmin", "nanargmax"]:
        v1 = getattr(np, attr)(x)
        v2 = getattr(x, attr)()
        v3 = getattr(np, attr)(np.asarray(x.copy()))
        print(attr, v1, v2, v3)
        assert(np.isclose(v1, v2))
        assert(np.isclose(v2, v3))

    x[m.nonzero()[1]] = np.inf

    for attr in ["isinf", "isnan", "isfinite"]:
        for w in [None, m]:
            print(attr, w)
            kwargs = {"where": w} if w is not None else {}
            m1 = getattr(x, attr)(**kwargs)
            if w is not None:
                kwargs["where"] = np.asarray(w.to_map(), dtype=bool)
                kwargs["out"] = np.zeros_like(kwargs["where"])
            m2 = maps.G3SkyMapMask(x, getattr(np, attr)(x, **kwargs).ravel())
            assert (m1 == m2).all()



#!/usr/bin/env python

from spt3g.core import G3FrameType
import unittest

class TestFramesTypes(unittest.TestCase):
    def test_key(self):
        self.assertEqual(G3FrameType.Map.key, 'M')

if __name__ == '__main__':
    unittest.main()



from .._libmaps import *

# Just run this, no symbols we need though
from .skymapaddons import *

from .azel import *
from .coordsysmodules import *
from .quathelpers import ang_to_quat, quat_to_ang, AddTimingToPointingQuats
from .map_modules import *
from .fitsio import *
from .maputils import *



from .. import core

@core.indexmod
class FixedLengthScans(object):
    '''Makes scans of length N timepoints.'''
    def __init__(self, N=1000):
        self.N = N
        self.count = 0
    def __call__(self, frame):
        ret = []
        if frame.type == core.G3FrameType.Timepoint:
            if self.count % self.N == 0:
                ret.append(core.G3Frame(core.G3FrameType.Scan))
            self.count += 1
        ret.append(frame)
        return ret

